"""Test PCC algorithm."""


from __future__ import annotations

from typing import Any

import pytest
import xarray as xr

from pycontrails import MetDataArray, MetDataset, Model
from pycontrails.models.humidity_scaling import ExponentialBoostLatitudeCorrectionHumidityScaling
from pycontrails.models.pcc import PCC


def test_pcc_default(met_pcc_pl: MetDataset, met_pcc_sl: MetDataset) -> None:
    """Test Default pcc algorithm."""
    with pytest.warns(UserWarning, match="originated from ECMWF"):
        PCC(met=met_pcc_pl, surface=met_pcc_sl)

    # little hack to avoid warning
    del met_pcc_pl.attrs["history"]
    met_pcc_pl.attrs["met_source"] = "not_ecmwf"

    model = PCC(met=met_pcc_pl, surface=met_pcc_sl)
    _pcc = model.eval()

    assert isinstance(_pcc, MetDataArray)
    assert isinstance(_pcc.data, xr.DataArray)
    assert isinstance(model.met, MetDataset)
    assert _pcc.name == "pcc"


@pytest.fixture(params=["Smith1990", "Sundqvist1989", "Slingo1980"])
def fancy_pcc(met_pcc_pl: MetDataset, met_pcc_sl: MetDataset, request: Any) -> Model:
    """Generate PCC models for each cloud model type."""
    params = {
        "cloud_model": request.param,
        "humidity_scaling": ExponentialBoostLatitudeCorrectionHumidityScaling(),
    }
    return PCC(met_pcc_pl, surface=met_pcc_sl, params=params)


def test_pcc_cloud_model(fancy_pcc: PCC) -> None:
    """Test PCC algorithm with alternate cloud models."""
    out = fancy_pcc.eval()
    assert isinstance(out, MetDataArray)
    assert isinstance(out.data, xr.DataArray)
    assert isinstance(fancy_pcc.met, MetDataset)
    assert out.name == "pcc"



"""Parameters for CoCiP uncertainty calculations."""

from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from typing import Any, ClassVar

import numpy as np
from scipy import stats
from scipy.stats.distributions import rv_frozen

from pycontrails.models.cocip.cocip_params import CocipParams
from pycontrails.models.humidity_scaling import ExponentialBoostHumidityScaling

logger = logging.getLogger(__name__)


class habit_dirichlet(rv_frozen):
    r"""Custom dirichlet distribution for habit weight distribution uncertainty.

    Scale the habit distributions by a dirichlet distribution
    with alpha parameter :math:`\alpha_{i} = 0.5 + \text{C} \text{G}_{i}`
    where :math:`\text{G}_{i}` is the approximate habit weight distributions
    defined in :attr:`CocipParams().habit_distributions`.

    References
    ----------
    - Table 2 in :cite:`schumannEffectiveRadiusIce2011`
    """

    def __init__(self, C: float = 96.0):
        self.C = C

    def rvs(self, *args: Any, **kwds: Any) -> np.ndarray:
        """Generate sample set of habit distributions.

        Sampled using dirichlet distribution.

        Parameters
        ----------
        *args
            Used to create a number of habit distributions
        **kwds
            Passed through to :func:`scipy.stats.dirichlet.rvs()`

        Returns
        -------
        np.ndarray
            Sampled habit weight distribution with the same shape
            as :attr:`CocipParams().habit_distributions`
        """
        if args or ("size" in kwds and kwds["size"] is not None and kwds["size"] > 1):
            raise ValueError("habit_dirichlet distribution only supports creating one rv at a time")

        default_distributions = CocipParams().habit_distributions
        alpha_i = [0.5 + self.C * G_i for G_i in default_distributions]
        distr = [stats.dirichlet(a) for a in alpha_i]

        habit_weights = default_distributions.copy()
        for i in range(1, habit_weights.shape[0]):
            habit_weights[i] = distr[i].rvs(**kwds)

        return habit_weights


@dataclass
class CocipUncertaintyParams(CocipParams):
    """Model parameters for CoCiP epistemic uncertainty.

    Any uncertainty parameters should end with the suffix `uncertainty`. See `__post_init__`.

    Uncertainty parameters take a `scipy.stats.distributions.rv_frozen` distribution.

    Default distributions have mean at CocipParam default values.

    To retain specific parameters as defaults, set the value of the key to None.

    Examples
    --------
    >>> import scipy.stats
    >>> from pycontrails.models.cocip import CocipUncertaintyParams

    # Override the `rhi_adj` field from `CocipParams` with a value in [0.97, 0.99]
    >>> distr = scipy.stats.uniform(loc=0.97, scale=0.02)
    >>> params = CocipUncertaintyParams(seed=123, rhi_adj_uncertainty=distr)
    >>> params.rhi_adj
    0.98364703

    # Once seeded, calling the class again gives a new value
    >>> params = CocipUncertaintyParams(rhi_adj_uncertainty=distr)
    >>> params.rhi_adj
    0.98310570

    >>> params = CocipUncertaintyParams(rf_lw_enhancement_factor_uncertainty=None)
    >>> params.rf_lw_enhancement_factor
    1.0
    """

    #: The random number generator is attached to the class (as opposed to an instance).
    #: If many instances of :class:`CocipUncertaintyParams` are needed, it is best to seed
    #: the random number generator once initially. Reseeding the generator will also
    #: reset it, giving rise to duplicate random numbers.
    rng: ClassVar[np.random.Generator] = np.random.default_rng(None)

    #: Reseed the random generator defined in ``__post_init__``
    seed: int | None = None

    # This might need fixing!
    # Whenever we overhaul / redo an uncertainty analysis, we may need to change
    # how it interacts with a humidity scaler
    humidity_scaling: ExponentialBoostHumidityScaling = ExponentialBoostHumidityScaling()

    #: Parameters for specific humidity and RHi enhancement
    #: This assumes Cocip.Params.humidity_scaling is an
    #: :class:`ExponentialBoostHumidityScaling`` instance.
    rhi_adj_uncertainty: rv_frozen | None = stats.norm(
        loc=ExponentialBoostHumidityScaling.default_params.rhi_adj, scale=0.1
    )
    rhi_boost_exponent_uncertainty: rv_frozen = stats.triang(
        loc=1.0,
        c=ExponentialBoostHumidityScaling.default_params.rhi_boost_exponent - 1.0,
        scale=1.0,
    )

    #: Schumann takes ``wind_shear_enhancement_exponent`` = 0.5 and discusses the case of 0 and 2/3
    #: as possibilities.
    #: With a value of 0, wind shear is not enhanced.
    wind_shear_enhancement_exponent_uncertainty: rv_frozen | None = stats.triang(
        loc=0.0, c=CocipParams.wind_shear_enhancement_exponent, scale=1.0
    )

    #: Schumann takes ``initial_wake_vortex_depth`` = 0.5 and discusses some
    #: uncertainty in this value. This parameter should be non-negative.
    initial_wake_vortex_depth_uncertainty: rv_frozen | None = stats.triang(
        loc=0.3, c=CocipParams.initial_wake_vortex_depth, scale=0.4
    )

    #: Schumann takes a default value of 0.1 and describes it as an "important adjustable parameter"
    #: Currently, `CocipParams` uses a default value of 0.5
    sedimentation_impact_factor_uncertainty: rv_frozen | None = stats.norm(
        loc=CocipParams.sedimentation_impact_factor, scale=0.1
    )

    #: Teoh 2022 (to appear) takes values between 70% decrease and 100% increase.
    #: This coincides with the log normal distribution defined below.
    nvpm_ei_n_enhancement_factor_uncertainty: rv_frozen | None = stats.lognorm(
        s=0.15, scale=1 / stats.lognorm(s=0.15).mean()
    )

    #: Scale shortwave radiative forcing.
    #: Table 2 in :cite:`schumannParametricRadiativeForcing2012`
    #: provides relative RMS error for SW/LW fit to the data generated
    #: by `libRadTran <http://www.libradtran.org/doku.php>`_.
    #: We use the average RMS error across all habit types (pg 1397) as the standard deviation
    #: of a normally distributed scaling factor for SW forcing
    rf_sw_enhancement_factor_uncertainty: rv_frozen | None = stats.norm(
        loc=CocipParams.rf_sw_enhancement_factor, scale=0.106
    )

    #: Scale longwave radiative forcing.
    #: Table 2 in :cite:`schumannParametricRadiativeForcing2012` provides relative error for SW/LW
    #: fit to the data generated by `libRadTran <http://www.libradtran.org/doku.php>`_.
    #: We use the average RMS error across all habit types (pg 1397) as the standard deviation
    #: of a normally distributed scaling factor for LW forcing.
    rf_lw_enhancement_factor_uncertainty: rv_frozen | None = stats.norm(
        loc=CocipParams.rf_lw_enhancement_factor, scale=0.071
    )

    #: Scale the habit distributions by a dirichlet distribution
    #: with alpha parameter :math:`\alpha_{i} = 0.5 + \text{C} \text{G}_{i}`
    #: where :math:`\text{G}_{i}` is the approximate habit weight distributions
    #: defined in :attr:`CocipParams().habit_distributions`.
    #: Higher values of :math:`\text{C}` correspond to higher confidence in initial estimates.
    habit_distributions_uncertainty: rv_frozen | None = habit_dirichlet(C=96)

    def __post_init__(self) -> None:
        """Override values of model parameters according to ranges."""
        if self.seed is not None:
            # Reset the class variable `rng`
            logger.info("Reset %s random seed to %s", self.__class__.__name__, self.seed)
            self.__class__.rng = np.random.default_rng(self.seed)

        # Override defaults value on `CocipParams` with random parameters
        for param, value in self.rvs().items():
            setattr(self, param, value)

    @property
    def uncertainty_params(self) -> dict[str, rv_frozen]:
        """Get dictionary of uncertainty parameters.

        Method checks for attributes ending in `"_uncertainty"`.

        Returns
        -------
        dict[str, dict[str, Any]]
            Uncertainty parameters and values
        """
        # handle these differently starting in version 0.27.0
        exclude = {"rhi_adj", "rhi_boost_exponent"}

        out = {}

        param_dict = asdict(self)
        for uncertainty_param, dist in param_dict.items():
            if uncertainty_param.endswith("_uncertainty") and dist is not None:
                param = uncertainty_param.split("_uncertainty")[0]
                if param not in exclude and param not in param_dict:
                    raise AttributeError(
                        f"Parameter {param} corresponding to uncertainty parameter "
                        f"{uncertainty_param} does not exist"
                    )

                if not isinstance(dist, rv_frozen):
                    raise AttributeError(
                        f"Uncertainty parameter '{uncertainty_param}' must be instance of "
                        "'scipy.stats.distributions.rv_frozen'"
                    )

                out[param] = dist

        return out

    def rvs(self, size: None | int = None) -> dict[str, float | np.ndarray]:
        """Call each distribution's `rvs` method to generate random parameters.

        Seed calls to `rvs` with class variable `rng`.

        Parameters
        ----------
        size : None | int, optional
            If specified, an `array` of values is generated for each uncertainty parameter.

        Returns
        -------
        dict[str, float | np.ndarray]
            Dictionary of random parameters. Dictionary keys consists of names of parameters in
            `CocipParams` to be overridden by random value.

        Examples
        --------
        >>> from pprint import pprint
        >>> from pycontrails.models.cocip import CocipUncertaintyParams
        >>> params = CocipUncertaintyParams(seed=456)
        >>> pprint(params.rvs())
        {'habit_distributions': array([[0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 0.0000000e+00,
               0.0000000e+00, 0.0000000e+00, 1.0000000e+00, 0.0000000e+00],
              [3.6062175e-04, 2.9157758e-01, 6.5901526e-04, 1.3415117e-03,
               5.0444747e-03, 2.6956137e-02, 6.7330402e-01, 7.5663364e-04],
              [2.3834489e-03, 3.6896354e-01, 1.3977488e-03, 4.7844747e-04,
               3.0474603e-01, 4.8765671e-03, 3.1715181e-01, 2.4031256e-06],
              [4.0944587e-03, 5.1772928e-01, 4.8383059e-05, 1.3166944e-03,
               1.0002965e-01, 3.7468293e-01, 3.9641498e-04, 1.7021718e-03],
              [1.5192166e-05, 3.8423434e-01, 5.0536448e-01, 9.4547205e-02,
               1.1378267e-03, 3.3812344e-03, 4.6305601e-03, 6.6891452e-03],
              [1.1620003e-03, 2.3161094e-03, 1.6705607e-04, 2.0091899e-02,
               9.7001791e-01, 1.5312615e-03, 3.9347797e-03, 7.7895907e-04]],
             dtype=float32),
        'initial_wake_vortex_depth': 0.4752143043559352,
        'nvpm_ei_n_enhancement_factor': 1.0094726146454185,
        'rf_lw_enhancement_factor': 0.9718354129386728,
        'rf_sw_enhancement_factor': 1.016450567639041,
        'rhi_adj': 1.0293074966694602,
        'rhi_boost_exponent': 1.3308280817638443,
        'sedimentation_impact_factor': 0.6433086271594898,
        'wind_shear_enhancement_exponent': 0.31735944665388505}
        """
        return {
            param: distr.rvs(size=size, random_state=self.rng)
            for param, distr in self.uncertainty_params.items()
        }

    def as_dict(self) -> dict[str, Any]:
        """Convert object to dictionary.

        Wrapper around :meth:`ModelBase.as_dict` that removes
        uncertainty specific parameters.

        Returns
        -------
        dict[str, Any]
            Dictionary version of self.
        """
        obj = super().as_dict()

        # remove seed and _uncertainty attributes
        # these will throw an error in `ModelBase._load_params()`
        keys_to_remove = ["seed"]
        for key in obj:
            if key.endswith("_uncertainty"):
                keys_to_remove.append(key)

        for key in keys_to_remove:
            obj.pop(key, None)

        return obj



"""Abstract classes for aircraft performance models."""

from __future__ import annotations

import abc
import dataclasses
from typing import Any, overload

from pycontrails.core.flight import Flight
from pycontrails.core.met import MetDataset
from pycontrails.core.models import Model, ModelParams
from pycontrails.core.vector import GeoVectorDataset


class AircraftPerformance(Model):
    """
    Support for standardizing aircraft performance methodologies.

    Currently just a container until additional models are implemented.
    """

    @overload
    def eval(self, source: Flight, **params: Any) -> Flight:
        ...

    # This is only included for type consistency with parent. This will raise.
    @overload
    def eval(self, source: None = None, **params: Any) -> Flight:
        ...

    @abc.abstractmethod
    def eval(self, source: Flight | None = None, **params: Any) -> Flight:
        """Evaluate the aircraft performance model."""


class AircraftPerformanceGrid(Model):
    """
    Support for standardizing aircraft performance methodologies on a grid.

    Currently just a container until additional models are implemented.
    """

    @overload
    def eval(self, source: GeoVectorDataset, **params: Any) -> GeoVectorDataset:
        ...

    @overload
    def eval(self, source: MetDataset | None = None, **params: Any) -> MetDataset:
        ...

    @abc.abstractmethod
    def eval(
        self, source: GeoVectorDataset | MetDataset | None = None, **params: Any
    ) -> GeoVectorDataset | MetDataset:
        """Evaluate the aircraft performance model."""


@dataclasses.dataclass
class AircraftPerformanceParams(ModelParams):
    """Parameters for :class:`AircraftPerformance`."""



"""Flight Data Handling."""

from __future__ import annotations

import dataclasses
import enum
import logging
import warnings
from typing import TYPE_CHECKING, Any

import numpy as np
import numpy.typing as npt
import pandas as pd
import pyproj
import scipy.signal
from overrides import overrides

from pycontrails.core.fuel import Fuel, JetA
from pycontrails.core.vector import AttrDict, GeoVectorDataset, VectorDataDict, VectorDataset
from pycontrails.physics import constants, geo, units

logger = logging.getLogger(__name__)

# optional imports
if TYPE_CHECKING:
    import matplotlib
    import traffic


class FlightPhase(enum.IntEnum):
    """Flight phase enumeration."""

    CLIMB = enum.auto()
    CRUISE = enum.auto()
    DESCENT = enum.auto()
    LEVEL_FLIGHT = enum.auto()
    NAN = enum.auto()


#: Max airport elevation, [:math:`ft`]
#: See `Daocheng_Yading_Airport <https://en.wikipedia.org/wiki/Daocheng_Yading_Airport>`_
MAX_AIRPORT_ELEVATION = 15_000.0

#: Min estimated cruise altitude, [:math:`ft`]
MIN_CRUISE_ALTITUDE = 20_000.0

#: Short haul duration cutoff, [:math:`s`]
SHORT_HAUL_DURATION = 3600.0

#: Set maximum speed compatible with "on_ground" indicator, [:math:`mph`]
#: Thresholds assessed based on scatter plot (150 knots = 278 km/h)
MAX_ON_GROUND_SPEED = 150.0


@dataclasses.dataclass
class Aircraft:
    """Base class for the physical parameters of the aircraft."""

    #: Aircraft type ICAO
    aircraft_type: str | None = None

    #: Aircraft type name
    aircraft_name: str | None = None

    #: Wingspan, [:math:`m`]
    wingspan: float | None = None

    #: Engine name
    engine_name: str | None = None

    #: Number of engines
    n_engine: int | None = None

    #: Maximum Mach number
    max_mach: float | None = None

    #: Maximum altitude, [:math:`m`]
    max_altitude: float | None = None


class Flight(GeoVectorDataset):
    """A single flight trajectory.

    Expect latitude-longitude coordinates in WGS 84.
    Expect altitude in [:math:`m`].
    Expect pressure level (`level`) in [:math:`hPa`].

    Use the attribute :attr:`attrs["crs"]` to specify coordinate reference system
    using `PROJ <https://proj.org/>`_ or `EPSG <https://epsg.org/home.html>`_ syntax.

    Parameters
    ----------
    data : dict[str, np.ndarray] | pd.DataFrame | VectorDataDict | VectorDataset | None
        Flight trajectory waypoints as data dictionary or :class:`pandas.DataFrame`.
        Must include columns ``time``, ``latitude``, ``longitude``, ``altitude`` or ``level``.
        Keyword arguments for ``time``, ``latitude``, ``longitude``, ``altitude`` or ``level``
        will override ``data`` inputs. Expects ``altitude`` in meters and ``time`` as a
        DatetimeLike (or array that can processed with :func:`pd.to_datetime`).
        Additional waypoint-specific data can be included as additional keys/columns.
    longitude : npt.ArrayLike, optional
        Flight trajectory waypoint longitude.
        Defaults to None.
    latitude : npt.ArrayLike, optional
        Flight trajectory waypoint latitude.
        Defaults to None.
    altitude : npt.ArrayLike, optional
        Flight trajectory waypoint altitude, [:math:`m`].
        Defaults to None.
    level : npt.ArrayLike, optional
        Flight trajectory waypoint pressure level, [:math:`hPa`].
        Defaults to None.
    time : npt.ArrayLike, optional
        Flight trajectory waypoint time.
        Defaults to None.
    attrs : dict[str, Any], optional
        Additional flight properties as a dictionary.
        Defaults to {}.
    copy : bool, optional
        Copy data on Flight creation.
        Defaults to True.
    aircraft : Aircraft, optional
        Aircraft flying this flight trajectory.
        Properties from the Aircraft will be added to :attr:`attrs`.
        Defaults to None.
    fuel : Fuel, optional
        Fuel used in flight trajectory.
        Defaults to JetA if None.
    drop_duplicated_times : bool, optional
        Drop duplicate times in flight trajectory. Defaults to False.
    **attrs_kwargs : Any
        Additional flight properties passed as keyword arguments.

    Raises
    ------
    KeyError
        Raises if ``data`` input does not contain at least ``time``, ``latitude``, ``longitude``,
        (``altitude`` or ``level``).

    Notes
    -----
    The `Traffic <https://traffic-viz.github.io/index.html>`_ library has many helpful
    flight processing utilities.

    See :class:`traffic.core.Flight` for more information.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from pycontrails import Flight

    >>> # Create `Flight` from a DataFrame.
    >>> df = pd.DataFrame({
    ...     "longitude": np.linspace(20, 30, 500),
    ...     "latitude": np.linspace(40, 10, 500),
    ...     "altitude": 10500,
    ...     "time": pd.date_range('2021-01-01T10', '2021-01-01T15', periods=500),
    ... })
    >>> fl = Flight(data=df, flight_id=123)  # specify a flight_id by keyword
    >>> fl
    Flight [4 keys x 500 length, 2 attributes]
    Keys: longitude, latitude, altitude, time
    Attributes:
    time                [2021-01-01 10:00:00, 2021-01-01 15:00:00]
    longitude           [20.0, 30.0]
    latitude            [10.0, 40.0]
    altitude            [10500.0, 10500.0]
    flight_id           123
    crs                 EPSG:4326

    >>> # Create `Flight` from keywords
    >>> fl = Flight(
    ...     longitude=np.linspace(20, 30, 200),
    ...     latitude=np.linspace(40, 30, 200),
    ...     altitude=11000 * np.ones(200),
    ...     time=pd.date_range('2021-01-01T12', '2021-01-01T14', periods=200),
    ... )
    >>> fl
    Flight [4 keys x 200 length, 1 attributes]
        Keys: longitude, latitude, time, altitude
        Attributes:
        time                [2021-01-01 12:00:00, 2021-01-01 14:00:00]
        longitude           [20.0, 30.0]
        latitude            [30.0, 40.0]
        altitude            [11000.0, 11000.0]
        crs                 EPSG:4326

    >>> # Access the underlying data as DataFrame
    >>> fl.dataframe.head()
       longitude   latitude                          time  altitude
    0  20.000000  40.000000 2021-01-01 12:00:00.000000000   11000.0
    1  20.050251  39.949749 2021-01-01 12:00:36.180904522   11000.0
    2  20.100503  39.899497 2021-01-01 12:01:12.361809045   11000.0
    3  20.150754  39.849246 2021-01-01 12:01:48.542713567   11000.0
    4  20.201005  39.798995 2021-01-01 12:02:24.723618090   11000.0
    """

    #: Fuel used in flight trajectory
    fuel: Fuel

    def __init__(
        self,
        data: dict[str, npt.ArrayLike]
        | pd.DataFrame
        | VectorDataDict
        | VectorDataset
        | None = None,
        longitude: npt.ArrayLike | None = None,
        latitude: npt.ArrayLike | None = None,
        altitude: npt.ArrayLike | None = None,
        time: npt.ArrayLike | None = None,
        attrs: dict[str, Any] | AttrDict | None = None,
        copy: bool = True,
        aircraft: Aircraft | None = None,
        fuel: Fuel | None = None,
        drop_duplicated_times: bool = False,
        **attrs_kwargs: Any,
    ) -> None:
        # set aircraft properties to attrs
        # Note input attrs will override aircraft properties
        # TODO: refactor aircraft to handle the same way as Fuel?
        if aircraft is not None:
            aircraft_attrs = dataclasses.asdict(aircraft)
            if attrs:
                aircraft_attrs.update(attrs)
                attrs = aircraft_attrs
            else:
                attrs = aircraft_attrs

        super().__init__(
            data=data,
            longitude=longitude,
            latitude=latitude,
            altitude=altitude,
            time=time,
            attrs=attrs,
            copy=copy,
            **attrs_kwargs,
        )

        # Set fuel - fuel instance is NOT copied
        self.fuel = fuel or JetA()

        # Check flight data for possible errors
        if np.any(self.altitude > 16000):
            flight_id = self.attrs.get("flight_id", "")
            flight_id = flight_id and f" for flight {flight_id}"
            warnings.warn(
                f"Flight altitude is high{flight_id}. Expected altitude unit is meters. "
                f"Found waypoint with altitude {self.altitude.max():.0f} m."
            )

        # Get time differences between waypoints
        if np.isnat(self["time"]).any():
            warnings.warn(
                "Flight trajectory contains NaT times. This will cause errors "
                "with segment-based methods (e.g. 'segment_true_airspeed')."
            )

        diff_ = np.diff(self["time"])

        # Ensure that time is sorted
        if self and np.any(diff_ < np.timedelta64(0)):
            if not copy:
                raise ValueError(
                    "`time` data must be sorted if `copy` is False on creation. "
                    "Set copy=False, or sort data before creating Flight."
                )
            warnings.warn("Sorting Flight data by `time`.")

            sorted_flight = self.sort("time")
            self.data = sorted_flight.data
            # Update diff_ ... we use it again below
            diff_ = np.diff(self["time"])

        # Check for duplicate times. If dropping duplicates,
        # keep the *first* occurrence of each time. This is achieved with
        # the np.insert (rather than np.append) function.
        filt = np.insert(diff_ > np.timedelta64(0), 0, True)
        if not np.all(filt):
            if drop_duplicated_times:
                filtered_flight = self.filter(filt, copy=False)
                self.data = filtered_flight.data
            else:
                warnings.warn(
                    "Flight contains duplicate times. This will cause errors "
                    "with segment-based methods (e.g. 'segment_true_airspeed'). Set "
                    "'drop_duplicated=True' or call the 'resample_and_fill' method."
                )

    @overrides
    def copy(self) -> Flight:
        return Flight(data=self.data, attrs=self.attrs, fuel=self.fuel, copy=True)

    @property
    def time_start(self) -> pd.Timestamp:
        """First waypoint time.

        Returns
        -------
        pd.Timestamp
            First waypoint time
        """
        return pd.Timestamp(np.nanmin(self["time"]))

    @property
    def time_end(self) -> pd.Timestamp:
        """Last waypoint time.

        Returns
        -------
        pd.Timestamp
            Last waypoint time
        """
        return pd.Timestamp(np.nanmax(self["time"]))

    @property
    def duration(self) -> pd.Timedelta:
        """Determine flight duration.

        Returns
        -------
        pd.Timedelta
            Difference between terminal and initial time
        """
        return pd.Timedelta(self.time_end - self.time_start)

    @property
    def max_time_gap(self) -> pd.Timedelta:
        """Return maximum time gap between waypoints along flight trajectory.

        Returns
        -------
        pd.Timedelta
            Gap size

        Examples
        --------
        >>> import numpy as np
        >>> fl = Flight(
        ...     longitude=np.linspace(20, 30, 200),
        ...     latitude=np.linspace(40, 30, 200),
        ...     altitude=11000 * np.ones(200),
        ...     time=pd.date_range('2021-01-01T12', '2021-01-01T14', periods=200),
        ... )
        >>> fl.max_time_gap
        Timedelta('0 days 00:00:36.180...')
        """
        return pd.Timedelta(np.nanmax(np.diff(self["time"])))

    @property
    def max_distance_gap(self) -> float:
        """Return maximum distance gap between waypoints along flight trajectory.

        Distance is calculated based on WGS84 geodesic.

        Returns
        -------
        float
            Maximum distance between waypoints, [:math:`m`]

        Raises
        ------
        NotImplementedError
            Raises when attr:`attrs["crs"]` is not EPSG:4326

        Examples
        --------
        >>> import numpy as np
        >>> fl = Flight(
        ...     longitude=np.linspace(20, 30, 200),
        ...     latitude=np.linspace(40, 30, 200),
        ...     altitude=11000 * np.ones(200),
        ...     time=pd.date_range('2021-01-01T12', '2021-01-01T14', periods=200),
        ... )
        >>> fl.max_distance_gap
        7391.27...
        """
        if self.attrs["crs"] != "EPSG:4326":
            raise NotImplementedError("Only implemented for EPSG:4326 CRS.")

        return self.segment_length()[:-1].max()

    @property
    def length(self) -> float:
        """Return flight length based on WGS84 geodesic.

        Returns
        -------
        float
            Total flight length, [:math:`m`]

        Raises
        ------
        NotImplementedError
            Raises when attr:`attrs["crs"]` is not EPSG:4326

        Examples
        --------
        >>> import numpy as np
        >>> fl = Flight(
        ...     longitude=np.linspace(20, 30, 200),
        ...     latitude=np.linspace(40, 30, 200),
        ...     altitude=11000 * np.ones(200),
        ...     time=pd.date_range('2021-01-01T12', '2021-01-01T14', periods=200),
        ... )
        >>> fl.length
        1436924.67...
        """
        if self.attrs["crs"] != "EPSG:4326":
            raise NotImplementedError("Only implemented for EPSG:4326 CRS.")

        # drop off the nan
        return np.nansum(self.segment_length()[:-1])

    # ------------
    # Segment Properties
    # ------------

    def segment_duration(self, dtype: np.dtype = np.dtype(np.float32)) -> npt.NDArray[np.float_]:
        r"""Compute time elapsed between waypoints in seconds.

        ``np.nan`` appended so the length of the output is the same as number of waypoints.

        Parameters
        ----------
        dtype : np.dtype
            Numpy dtype for time difference.
            Defaults to ``np.float64``

        Returns
        -------
        npt.NDArray[np.float_]
            Time difference between waypoints, [:math:`s`].
            Returns an array with dtype specified by``dtype``
        """

        return segment_duration(self.data["time"], dtype=dtype)

    def segment_length(self) -> npt.NDArray[np.float_]:
        """Compute spherical distance between flight waypoints.

        Helper function used in :meth:`length` and :meth:`length_met`.
        `np.nan` appended so the length of the output is the same as number of waypoints.

        Returns
        -------
        npt.NDArray[np.float_]
            Array of distances in [:math:`m`] between waypoints

        Raises
        ------
        NotImplementedError
            Raises when attr:`attrs["crs"]` is not EPSG:4326

        Examples
        --------
        >>> from pycontrails import Flight
        >>> fl = Flight(
        ... longitude=np.array([1, 2, 3, 5, 8]),
        ... latitude=np.arange(5),
        ... altitude=np.full(shape=(5,), fill_value=11000),
        ... time=pd.date_range('2021-01-01T12', '2021-01-01T14', periods=5),
        ... )
        >>> fl.segment_length()
        array([157255.03346286, 157231.08336815, 248456.48781503, 351047.44358851,
                           nan])

        See Also
        --------
        :func:`segment_length`
        """
        if self.attrs["crs"] != "EPSG:4326":
            raise NotImplementedError("Only implemented for EPSG:4326 CRS.")

        return geo.segment_length(self["longitude"], self["latitude"], self.altitude)

    def segment_angle(self) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        """Calculate sin/cos for the angle between each segment and the longitudinal axis.

        This is different from the usual navigational angle between two points known as *bearing*.

        *Bearing* in 3D spherical coordinates is referred to as *azimuth*.
        ::

                    (lon_2, lat_2)  X
                                   /|
                                  / |
                                 /  |
                                /   |
                               /    |
                              /     |
                             /      |
            (lon_1, lat_1)  X -------> longitude (x-axis)

        Returns
        -------
        npt.NDArray[np.float_], npt.NDArray[np.float_]
            sin(a), cos(a), where ``a`` is the angle between the segment and the longitudinal axis
            The final values are of both arrays are ``np.nan``.

        See Also
        --------
        :func:`geo.segment_angle`
        :func:`units.heading_to_longitudinal_angle`
        :meth:`segment_azimuth`
        :func:`geo.forward_azimuth`

        Examples
        --------
        >>> from pycontrails import Flight
        >>> fl = Flight(
        ... longitude=np.array([1, 2, 3, 5, 8]),
        ... latitude=np.arange(5),
        ... altitude=np.full(shape=(5,), fill_value=11000),
        ... time=pd.date_range('2021-01-01T12', '2021-01-01T14', periods=5),
        ... )
        >>> sin, cos = fl.segment_angle()
        >>> sin
        array([0.70716063, 0.70737598, 0.44819424, 0.31820671,        nan])

        >>> cos
        array([0.70705293, 0.70683748, 0.8939362 , 0.94802136,        nan])

        """
        return geo.segment_angle(self["longitude"], self["latitude"])

    def segment_azimuth(self) -> npt.NDArray[np.float_]:
        """Calculate (forward) azimuth at each waypoint.

        Method calls `pyproj.Geod.inv`, which is slow. See `geo.forward_azimuth`
        for an outline of a faster implementation.

        .. versionchanged:: 0.33.7

            The dtype of the output now matches the dtype of ``self["longitude"]``.

        Returns
        -------
        npt.NDArray[np.float_]
            Array of azimuths.
        """
        lon = self["longitude"]
        lat = self["latitude"]

        lons1 = lon[:-1]
        lats1 = lat[:-1]
        lons2 = lon[1:]
        lats2 = lat[1:]

        geod = pyproj.Geod(a=constants.radius_earth)
        az, *_ = geod.inv(lons1, lats1, lons2, lats2)

        # NOTE: geod.inv automatically promotes to float64. We match the dtype of lon.
        out = np.empty_like(lon)
        out[:-1] = az
        # Convention: append nan
        out[-1] = np.nan

        return out

    def segment_groundspeed(
        self, smooth: bool = False, window_length: int = 7, polyorder: int = 1
    ) -> npt.NDArray[np.float_]:
        """Return groundspeed across segments.

        Calculate by dividing the horizontal segment length by the difference in waypoint times.

        Parameters
        ----------
        smooth : bool, optional
            Smooth airspeed with Savitzky-Golay filter.
            Defaults to False.
        window_length : int, optional
            Passed directly to :func:`scipy.signal.savgol_filter`, by default 7.
        polyorder : int, optional
            Passed directly to :func:`scipy.signal.savgol_filter`, by default 1.

        Returns
        -------
        npt.NDArray[np.float_]
            Groundspeed of the segment, [:math:`m s^{-1}`]
        """
        # get horizontal distance - set altitude to 0
        horizontal_segment_length = geo.segment_haversine(self["longitude"], self["latitude"])

        # time between waypoints, in seconds
        dt_sec = np.empty_like(horizontal_segment_length)
        dt_sec[:-1] = np.diff(self["time"]) / np.timedelta64(1, "s")
        dt_sec[-1] = np.nan

        # calculate groundspeed
        groundspeed = horizontal_segment_length / dt_sec

        # Savitzky-Golay filter
        if smooth:
            # omit final nan value, then reattach it afterwards
            groundspeed[:-1] = _sg_filter(groundspeed[:-1], window_length, polyorder)
            groundspeed[-1] = np.nan

        return groundspeed

    def segment_true_airspeed(
        self,
        u_wind: npt.NDArray[np.float_] | None = None,
        v_wind: npt.NDArray[np.float_] | None = None,
        smooth: bool = True,
        window_length: int = 7,
        polyorder: int = 1,
    ) -> npt.NDArray[np.float_]:
        r"""Calculate the true airspeed [:math:`m/s`] from the ground speed and horizontal winds.

        The calculated ground speed will first be smoothed with a Savitzky-Golay filter if enabled.

        Parameters
        ----------
        u_wind : npt.NDArray[np.float_], optional
            U wind speed, [:math:`m \ s^{-1}`].
            Defaults to 0 for all waypoints.
        v_wind : npt.NDArray[np.float_], optional
            V wind speed, [:math:`m \ s^{-1}`].
            Defaults to 0 for all waypoints.
        smooth : bool, optional
            Smooth airspeed with Savitzky-Golay filter.
            Defaults to True.
        window_length : int, optional
            Passed directly to :func:`scipy.signal.savgol_filter`, by default 7.
        polyorder : int, optional
            Passed directly to :func:`scipy.signal.savgol_filter`, by default 1.

        Returns
        -------
        npt.NDArray[np.float_]
            True wind speed of each segment, [:math:`m \ s^{-1}`]
        """
        groundspeed = self.segment_groundspeed(smooth, window_length, polyorder)

        if u_wind is None:
            u_wind = np.zeros_like(groundspeed)

        if v_wind is None:
            v_wind = np.zeros_like(groundspeed)

        sin_a, cos_a = self.segment_angle()
        gs_x = groundspeed * cos_a
        gs_y = groundspeed * sin_a
        tas_x = gs_x - u_wind
        tas_y = gs_y - v_wind

        return np.sqrt(tas_x * tas_x + tas_y * tas_y)

    def segment_mach_number(
        self, true_airspeed: npt.NDArray[np.float_], air_temperature: npt.NDArray[np.float_]
    ) -> npt.NDArray[np.float_]:
        r"""Calculate the mach number of each segment.

        Parameters
        ----------
        true_airspeed : npt.NDArray[np.float_]
            True airspeed of the segment, [:math:`m \ s^{-1}`].
            See :meth:`segment_true_airspeed`.
        air_temperature : npt.NDArray[np.float_]
            Average air temperature of each segment, [:math:`K`]

        Returns
        -------
        npt.NDArray[np.float_]
            Mach number of each segment
        """
        return units.tas_to_mach_number(true_airspeed, air_temperature)

    def segment_rocd(self) -> npt.NDArray[np.float_]:
        """Calculate the rate of climb and descent (ROCD).

        Returns
        -------
        npt.NDArray[np.float_]
            Rate of climb and descent over segment, [:math:`ft min^{-1}`]

        See Also
        --------
        :func:`segment_rocd`
        """
        return segment_rocd(self.segment_duration(), self.altitude_ft)

    def segment_phase(
        self,
        threshold_rocd: float = 250.0,
        min_cruise_altitude_ft: float = 20000.0,
    ) -> npt.NDArray[np.uint8]:
        """Identify the phase of flight (climb, cruise, descent) for each segment.

        Parameters
        ----------
        threshold_rocd : float, optional
            ROCD threshold to identify climb and descent, [:math:`ft min^{-1}`].
            Currently set to 250 ft/min.
        min_cruise_altitude_ft : float, optional
            Minimum altitude for cruise, [:math:`ft`]
            This is specific for each aircraft type,
            and can be approximated as 50% of the altitude ceiling.
            Defaults to 20000 ft.

        Returns
        -------
        npt.NDArray[np.uint8]
            Array of values enumerating the flight phase.
            See :attr:`flight.FlightPhase` for enumeration.

        See Also
        --------
        :attr:`FlightPhase`
        :func:`segment_phase`
        :func:`segment_rocd`
        """
        return segment_phase(
            self.segment_rocd(),
            self.altitude_ft,
            threshold_rocd=threshold_rocd,
            min_cruise_altitude_ft=min_cruise_altitude_ft,
        )

    # ------------
    # Filter/Resample
    # ------------

    def filter_by_first(self) -> Flight:
        """Keep first row of group of waypoints with identical coordinates.

        Chaining this method with `resample_and_fill` often gives a cleaner trajectory
        when using noisy flight waypoints.

        Returns
        -------
        Flight
            Filtered Flight instance

        Examples
        --------
        >>> from datetime import datetime
        >>> import pandas as pd

        >>> df = pd.DataFrame()
        >>> df['longitude'] = [0, 0, 50]
        >>> df['latitude'] = 0
        >>> df['altitude'] = 0
        >>> df['time'] = [datetime(2020, 1, 1, h) for h in range(3)]

        >>> fl = Flight(df)

        >>> fl.filter_by_first().dataframe
           longitude  latitude  altitude                time
        0        0.0       0.0       0.0 2020-01-01 00:00:00
        1       50.0       0.0       0.0 2020-01-01 02:00:00
        """
        df = self.dataframe.groupby(["longitude", "latitude"], sort=False).first().reset_index()
        return Flight(data=df, attrs=self.attrs)

    def resample_and_fill(
        self,
        freq: str = "1T",
        fill_method: str = "geodesic",
        geodesic_threshold: float = 100e3,
        nominal_rocd: float = constants.nominal_rocd,
        drop: bool = True,
    ) -> Flight:
        """Resample and fill flight trajectory with geodesics and linear interpolation.

        Waypoints are resampled according to the frequency `freq`. Values for :attr:`data` columns
        `longitude`, `latitude`, and `altitude` are interpolated.

        Parameters
        ----------
        freq : str, optional
            Resampling frequency, by default "1T"
        fill_method : {"geodesic", "linear"}, optional
            Choose between ``"geodesic"`` and ``"linear"``, by default ``"geodesic"``.
            In geodesic mode, large gaps between waypoints are filled with geodesic
            interpolation and small gaps are filled with linear interpolation. In linear
            mode, all gaps are filled with linear interpolation.
        geodesic_threshold : float, optional
            Threshold for geodesic interpolation, [:math:`m`].
            If the distance between consecutive waypoints is under this threshold,
            values are interpolated linearly.
        nominal_rocd : float | None, optional
            Nominal rate of climb / descent for aircraft type.
            Defaults to :attr:`constants.nominal_rocd`.
        drop : bool, optional
            Drop any columns that are not resampled and filled.
            Defaults to ``True``, dropping all keys outside of "time", "latitude",
            "longitude" and "altitude". If set to False, the extra keys will be
            kept but filled with ``nan`` or ``None`` values, depending on the data type.

        Returns
        -------
        Flight
            Filled Flight

        Raises
        ------
        ValueError
            Unknown `fill_method`

        Examples
        --------
        >>> from datetime import datetime
        >>> import pandas as pd

        >>> df = pd.DataFrame()
        >>> df['longitude'] = [0, 0, 50]
        >>> df['latitude'] = 0
        >>> df['altitude'] = 0
        >>> df['time'] = [datetime(2020, 1, 1, h) for h in range(3)]

        >>> fl = Flight(df)
        >>> fl.dataframe
           longitude  latitude  altitude                time
                   0        0.0       0.0       0.0 2020-01-01 00:00:00
                   1        0.0       0.0       0.0 2020-01-01 01:00:00
                   2       50.0       0.0       0.0 2020-01-01 02:00:00

        >>> fl.resample_and_fill('10T').dataframe  # resample with 10 minute frequency
                          time  longitude  latitude  altitude
        0  2020-01-01 00:00:00   0.000000       0.0       0.0
        1  2020-01-01 00:10:00   0.000000       0.0       0.0
        2  2020-01-01 00:20:00   0.000000       0.0       0.0
        3  2020-01-01 00:30:00   0.000000       0.0       0.0
        4  2020-01-01 00:40:00   0.000000       0.0       0.0
        5  2020-01-01 00:50:00   0.000000       0.0       0.0
        6  2020-01-01 01:00:00   0.000000       0.0       0.0
        7  2020-01-01 01:10:00   8.928571       0.0       0.0
        8  2020-01-01 01:20:00  16.964286       0.0       0.0
        9  2020-01-01 01:30:00  25.892857       0.0       0.0
        10 2020-01-01 01:40:00  33.928571       0.0       0.0
        11 2020-01-01 01:50:00  41.964286       0.0       0.0
        12 2020-01-01 02:00:00  50.000000       0.0       0.0
        """
        methods = "geodesic", "linear"
        if fill_method not in methods:
            raise ValueError(f'Unknown `fill_method`. Supported  methods: {", ".join(methods)}')

        # STEP 1: Prepare DataFrame on which we'll perform resampling
        df = self.to_dataframe()

        # put altitude on dataframe if its not already there
        if "altitude" not in df:
            df["altitude"] = self.altitude

        # always drop level
        if "level" in df:
            df.drop(columns="level", inplace=True)

        # drop all cols except time/lon/lat/alt
        if drop:
            df = df.loc[:, ["time", "longitude", "latitude", "altitude"]]

        # STEP 2: Fill large horizontal gaps with interpolated geodesics
        if fill_method == "geodesic":
            filled = self._geodesic_interpolation(geodesic_threshold)
            if filled is not None:
                df = pd.concat([df, filled])

        # STEP 3: Set the time index, and sort it
        df = df.set_index("time").sort_index()

        # STEP 4: Some adhoc code for dealing with antimeridian.
        # Idea: A flight likel crosses the antimeridian if
        #   `min_pos > 90` and `max_neg < -90`
        # This is not foolproof: it assumes the full trajectory will not
        # span more than 180 longitude degrees. There could be flights that
        # violate this near the poles (but this would be very rare -- flights
        # would instead wrap the other way). For this flights spanning the
        # antimeridian, we translate them to a common "chart" away from the
        # antimeridian (see variable `shift`), then apply the interpolation,
        # then shift back to their original position.
        lon = df["longitude"].to_numpy()
        sign_ = np.sign(lon)
        min_pos = np.min(lon[sign_ == 1], initial=np.inf)
        max_neg = np.max(lon[sign_ == -1], initial=-np.inf)
        if (180 - min_pos) + (180 + max_neg) < 180 and min_pos < np.inf and max_neg > -np.inf:
            # In this case, we believe the flight crosses the antimeridian
            shift = min_pos
            # So we shift the longitude "chart"
            df["longitude"] = (df["longitude"] - shift) % 360
        else:
            shift = None

        # STEP 5: Resample flight to freq
        df = df.resample(freq).first()

        # STEP 6: Linearly interpolate small horizontal gaps and account
        # for previous longitude shift.
        keys = ["latitude", "longitude"]
        df.loc[:, keys] = df.loc[:, keys].interpolate(method="linear")
        if shift is not None:
            # We need to translate back to the original chart here
            df["longitude"] += shift
            df["longitude"] = ((df["longitude"] + 180) % 360) - 180

        # STEP 7: Interpolate nan values in altitude
        if df["altitude"].isna().any():
            df_freq = df.index.freq.delta.to_numpy()
            new_alt = _altitude_interpolation(df["altitude"].to_numpy(), nominal_rocd, df_freq)
            _verify_altitude(new_alt, nominal_rocd, df_freq)
            df["altitude"] = new_alt

        # finally reset index
        df = df.reset_index()
        return Flight(data=df, attrs=self.attrs)

    def _geodesic_interpolation(self, geodesic_threshold: float) -> pd.DataFrame | None:
        """Geodesic interpolate between large gaps between waypoints.

        Parameters
        ----------
        geodesic_threshold : float
            The threshold for large gap, [:math:`m`].

        Returns
        -------
        pd.DataFrame | None
            Generated waypoints to be merged into underlying :attr:`data`.
            Return `None` if no new waypoints are created.

        Raises
        ------
        NotImplementedError
            Raises when attr:`attrs["crs"]` is not EPSG:4326
        """
        if self.attrs["crs"] != "EPSG:4326":
            raise NotImplementedError("Only implemented for EPSG:4326 CRS.")

        # Omit the final nan and ensure index + 1 (below) is well defined
        segs = self.segment_length()[:-1]

        # For default geodesic_threshold, we expect gap_indices to be very
        # sparse (so the for loop below is cheap)
        gap_indices = np.nonzero(segs > geodesic_threshold)[0]
        if not np.any(gap_indices):
            # For most flights, gap_indices is empty. It's more performant
            # to exit now rather than build an empty DataFrame below.
            return None

        geod = pyproj.Geod(ellps="WGS84")
        longitudes: list[float] = []
        latitudes: list[float] = []
        times: list[np.ndarray] = []

        for index in gap_indices:
            lon0, lat0, t0 = (
                self["longitude"][index],
                self["latitude"][index],
                self["time"][index],
            )
            lon1, lat1, t1 = (
                self["longitude"][index + 1],
                self["latitude"][index + 1],
                self["time"][index + 1],
            )
            distance = segs[index]
            n_steps = distance // geodesic_threshold  # number of new waypoints to generate

            # This is the expensive call within the for-loop
            # NOTE: geod.npts does not return the initial or terminal points
            lonlats: list[tuple[float, float]] = geod.npts(lon0, lat0, lon1, lat1, n_steps)

            lons, lats = zip(*lonlats)
            longitudes.extend(lons)
            latitudes.extend(lats)

            # + 1 to denominator to stay consistent with geod.npts (only interior points)
            t_step = (t1 - t0) / (n_steps + 1)
            # substract 0.5 * t_step to ensure round-off error doesn't put final arange point
            # very close to t1
            t_range = np.arange(t0 + t_step, t1 - 0.5 * t_step, t_step)
            times.append(t_range)

        times_ = np.concatenate(times)
        return pd.DataFrame({"longitude": longitudes, "latitude": latitudes, "time": times_})

    # ------------
    # I / O
    # ------------

    def to_geojson_linestring(self) -> dict[str, Any]:
        """Return trajectory as geojson FeatureCollection containing single LineString.

        Returns
        -------
        dict[str, Any]
            Python representation of geojson FeatureCollection
        """
        points = _return_linestring(
            {
                "longitude": self["longitude"],
                "latitude": self["latitude"],
                "altitude": self.altitude,
            }
        )
        geometry = {"type": "LineString", "coordinates": points}
        properties = {
            "start_time": self.time_start.isoformat(),
            "end_time": self.time_end.isoformat(),
        }
        properties.update(self.constants)
        linestring = {"type": "Feature", "geometry": geometry, "properties": properties}

        return {"type": "FeatureCollection", "features": [linestring]}

    def to_geojson_multilinestring(
        self, key: str, split_antimeridian: bool = True
    ) -> dict[str, Any]:
        """Return trajectory as GeoJSON FeatureCollection of MultiLineStrings.

        Flight :attr:`data` is grouped according to values of ``key``. Each group gives rise to a
        Feature containing a MultiLineString geometry. LineStrings can be split over the
        antimeridian.

        Parameters
        ----------
        key : str
            Name of :attr:`data` column to group by
        split_antimeridian : bool, optional
            Split linestrings that cross the antimeridian.
            Defaults to True

        Returns
        -------
        dict[str, Any]
            Python representation of GeoJSON FeatureCollection of MultiLinestring Features

        Raises
        ------
        KeyError
            :attr:`data` does not contain column ``key``
        """
        if key not in self.dataframe.columns:
            raise KeyError(f"Column {key} does not exist in data.")

        jump_index = _antimeridian_index(pd.Series(self["longitude"]), self.attrs["crs"])

        def _group_to_feature(group: pd.DataFrame) -> dict[str, str | dict[str, Any]]:
            subgrouping = group.index.to_series().diff().ne(1).cumsum()
            # additional splitting at antimeridian
            if jump_index in subgrouping and split_antimeridian:
                subgrouping.loc[jump_index:] += 1
            multi_ls = [_return_linestring(g) for _, g in group.groupby(subgrouping)]
            geometry = {"type": "MultiLineString", "coordinates": multi_ls}

            # adding in static properties
            properties: dict[str, Any] = {key: group.name}
            properties.update(self.constants)
            return {"type": "Feature", "geometry": geometry, "properties": properties}

        features = self.dataframe.groupby(key).apply(_group_to_feature).values.tolist()
        return {"type": "FeatureCollection", "features": features}

    def to_traffic(self) -> "traffic.core.Flight":
        """Convert Flight instance to :class:`traffic.core.Flight` instance.

        See https://traffic-viz.github.io/traffic.core.flight.html#traffic.core.Flight

        Returns
        -------
        :class:`traffic.core.Flight`
            `traffic.core.Flight` instance

        Raises
        ------
        ModuleNotFoundError
            `traffic` package not installed
        """
        try:
            import traffic.core
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "This requires the 'traffic' module, which can be installed using "
                "'pip install traffic'. See the installation documentation at "
                "https://traffic-viz.github.io/installation.html for more information."
            ) from e

        return traffic.core.Flight(
            self.to_dataframe(copy=True).rename(columns={"time": "timestamp"})
        )

    # ------------
    # MET
    # ------------

    def length_met(self, key: str, threshold: float = 1.0) -> float:
        """Calculate total horizontal distance where column ``key`` exceeds``threshold``.

        Parameters
        ----------
        key : str
            Column key in :attr:`data`
        threshold : float
            Consider trajectory waypoints whose associated ``key`` value exceeds ``threshold``,
            by default 1.0

        Returns
        -------
        float
            Total distance, [:math:`m`]

        Raises
        ------
        KeyError
            :attr:`data` does not contain column ``key``
        NotImplementedError
            Raises when attr:`attrs["crs"]` is not EPSG:4326

        Examples
        --------
        >>> from datetime import datetime
        >>> import pandas as pd
        >>> import numpy as np
        >>> from pycontrails.datalib.ecmwf import ERA5
        >>> from pycontrails import Flight

        >>> # Get met data
        >>> times = (datetime(2022, 3, 1, 0),  datetime(2022, 3, 1, 3))
        >>> variables = ["air_temperature", "specific_humidity"]
        >>> levels = [300, 250, 200]
        >>> era5 = ERA5(time=times, variables=variables, pressure_levels=levels)
        >>> met = era5.open_metdataset(xr_kwargs=dict(parallel=False))

        >>> # Build flight
        >>> df = pd.DataFrame()
        >>> df['time'] = pd.date_range('2022-03-01T00', '2022-03-01T03', periods=11)
        >>> df['longitude'] = np.linspace(-20, 20, 11)
        >>> df['latitude'] = np.linspace(-20, 20, 11)
        >>> df['altitude'] = np.linspace(9500, 10000, 11)
        >>> fl = Flight(df).resample_and_fill('10S')

        >>> # Intersect and attach
        >>> fl["air_temperature"] = fl.intersect_met(met['air_temperature'])
        >>> fl["air_temperature"]
        array([235.94658, 235.55774, 235.56766, ..., 234.59956, 234.60406,
               234.60846], dtype=float32)

        >>> # Length (in meters) of waypoints whose temperature exceeds 236K
        >>> fl.length_met("air_temperature", threshold=236)
        3587431.887...

        >>> # Proportion (with respect to distance) of waypoints whose temperature exceeds 236K
        >>> fl.proportion_met("air_temperature", threshold=236)
        0.576076...
        """
        if key not in self.data:
            raise KeyError(f"Column {key} does not exist in data.")
        if self.attrs["crs"] != "EPSG:4326":
            raise NotImplementedError("Only implemented for EPSG:4326 CRS.")

        # The column of interest may contain floating point values less than 1.
        # In this case, if the default threshold is not changed, warn the user that the behavior
        # might not be what is expected.

        # Checking if column of interest contains floating point values below 1
        if threshold == 1.0 and ((self[key] > 0) & (self[key] < 1)).any():
            warnings.warn(
                f"Column {key} contains real numbers between 0 and 1. "
                "To include these values in this calculation, change the `threshold` parameter "
                "or modify the underlying DataFrame in place."
            )

        segs = self.segment_length()[:-1]  # lengths between waypoints, dropping off the nan

        # giving each waypoint the average of the segments on either side side
        segs = np.concatenate([segs[:1], (segs[1:] + segs[:-1]) / 2, segs[-1:]])

        # filter by region of interest
        indices = np.where(self[key] >= threshold)[0]

        return np.sum(segs[indices])

    def proportion_met(self, key: str, threshold: float = 1.0) -> float:
        """Calculate proportion of flight with certain meteorological constraint.

        Parameters
        ----------
        key : str
            Column key in :attr:`data`
        threshold : float
            Consider trajectory waypoints whose associated ``key`` value exceeds ``threshold``,
            Defaults to 1.0

        Returns
        -------
        float
            Ratio
        """
        try:
            return self.length_met(key, threshold) / self.length
        except ZeroDivisionError:
            return 0.0

    # ------------
    # Visualization
    # ------------

    def plot(self, **kwargs: Any) -> "matplotlib.axes.Axes":
        """Plot flight trajectory longitude-latitude values.

        Parameters
        ----------
        **kwargs : Any
            Additional plot properties to passed to `pd.DataFrame.plot`

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Plot
        """
        ax = self.dataframe.plot(x="longitude", y="latitude", legend=False, **kwargs)
        ax.set(xlabel="longitude", ylabel="latitude")
        return ax


def _return_linestring(data: dict[str, npt.NDArray[np.float_]]) -> list[list[float]]:
    """Return list of coordinates for geojson constructions.

    Parameters
    ----------
    data : dict[str, npt.NDArray[np.float_]]
        :attr:`data` containing `longitude`, `latitude`, and `altitude` keys

    Returns
    -------
    list[list[float]]
        The list of coordinates
    """
    # rounding to reduce the size of resultant json arrays
    points = zip(  # pylint: disable=zip-builtin-not-iterating
        np.around(data["longitude"], decimals=4),
        np.around(data["latitude"], decimals=4),
        np.around(data["altitude"], decimals=4),
    )
    return [list(p) for p in points]


def _antimeridian_index(longitude: pd.Series, crs: str = "EPSG:4326") -> int:
    """Return index after flight crosses antimeridian, or -1 if flight does not cross.

    Parameters
    ----------
    longitude : pd.Series
        longitude values with an integer index
    crs : str, optional
        Coordinate Reference system for longitude specified in EPSG format.
        Currently only supports "EPSG:4326" and "EPSG:3857".

    Returns
    -------
    int
        Index after jump or -1

    Raises
    ------
    ValueError
        CRS is not supported.
        Flight crosses antimeridian several times.
    """
    # FIXME: This logic here is somewhat outdated - the _interpolate_altitude
    # method handles this somewhat more reliably
    # This function should get updated to follow the logic there.
    # WGS84
    if crs in ["EPSG:4326"]:
        l1 = (-180.0, -90.0)
        l2 = (90.0, 180.0)

    # pseudo mercator
    elif crs in ["EPSG:3857"]:
        # values calculated through pyproj.Transformer
        l1 = (-20037508.342789244, -10018754.171394622)
        l2 = (10018754.171394622, 20037508.342789244)

    else:
        raise ValueError("CRS must be one of EPSG:4326 or EPSG:3857")

    # TODO: When nans exist, this method *may* not find the meridian
    if np.any(np.isnan(longitude)):
        warnings.warn("Anti-meridian index can't be found accurately with nan values in longitude")

    s1 = longitude.between(*l1)
    s2 = longitude.between(*l2)
    jump12 = longitude[s1 & s2.shift()]
    jump21 = longitude[s1.shift() & s2]
    jump_index = pd.concat([jump12, jump21]).index.to_list()

    if len(jump_index) > 1:
        raise ValueError("Only implemented for trajectories jumping the antimeridian at most once.")
    if len(jump_index) == 1:
        return jump_index[0]

    return -1


def _sg_filter(
    vals: npt.NDArray[np.float_], window_length: int = 7, polyorder: int = 1
) -> npt.NDArray[np.float_]:
    """Apply Savitzky-Golay filter to smooth out noise in the time-series data.

    Used to smooth true airspeed, fuel flow, and altitude.

    Parameters
    ----------
    vals : npt.NDArray[np.float_]
        Input array
    window_length : int, optional
        Parameter for :func:`scipy.signal.savgol_filter`
    polyorder : int, optional
        Parameter for :func:`scipy.signal.savgol_filter`

    Returns
    -------
    npt.NDArray[np.float_]
        Smoothed values

    Raises
    ------
    ArithmeticError
        Raised if NaN values input to SG filter
    """
    # The window_length must be less than or equal to the number of data points available.
    window_length = min(window_length, vals.size)

    # The time window_length must be odd.
    if (window_length % 2) == 0:
        window_length -= 1

    # If there is not enough data points to perform smoothing, return the mean ground speed
    if window_length <= polyorder:
        return np.full_like(vals, np.nanmean(vals))

    if np.isnan(vals).any():
        raise ArithmeticError("NaN values not supported by SG filter.")

    return scipy.signal.savgol_filter(vals, window_length, polyorder)


def _altitude_interpolation(
    altitude: npt.NDArray[np.float_], nominal_rocd: float, freq: np.timedelta64
) -> npt.NDArray[np.float_]:
    """Interpolate nan values in `altitude` array.

    Suppose each group of consecutive nan values is enclosed by `a0` and `a1` with
    corresponding time values `t0` and `t1` respectively. This function immediately
    climbs or descends starting at `t0` (depending on the sign of `a1 - a0`). Once
    the filled altitude values reach the terminal value `a1`, the remaining nans
    are filled with `a1.

    Parameters
    ----------
    altitude : npt.NDArray[np.float_]
        Array of altitude values containing nan values. This function will raise
        an error if `altitude` does not contain nan values. Moreover, this function
        assumes the initial and final entries in `altitude` are not nan.
    nominal_rocd : float
        Nominal rate of climb/descent, in m/s
    freq : np.timedelta64
        Frequency of time index associated to `altitude`.

    Returns
    -------
    npt.NDArray[np.float_]
        Altitude after nan values have been filled
    """
    # Determine nan state of altitude
    isna = np.isnan(altitude)
    start_na = ~isna[:-1] & isna[1:]
    start_na = np.append(start_na, False)
    end_na = isna[:-1] & ~isna[1:]
    end_na = np.insert(end_na, 0, False)

    # And get the size of each group of consecutive nan values
    start_na_idxs = start_na.nonzero()[0]
    end_na_idxs = end_na.nonzero()[0]
    na_group_size = end_na_idxs - start_na_idxs

    # Form array of cumulative altitude values if the flight were to climb
    # at nominal_rocd over each group of nan
    cumalt_list = [np.arange(1, size, dtype=float) for size in na_group_size]
    cumalt = np.concatenate(cumalt_list)
    cumalt = cumalt * nominal_rocd * freq / np.timedelta64(1, "s")

    # Expand cumalt to the full size of altitude
    nominal_fill = np.zeros_like(altitude)
    nominal_fill[isna] = cumalt

    # Use pandas to forward and backfill altitude values
    s = pd.Series(altitude)
    s_ff = s.fillna(method="ffill")
    s_bf = s.fillna(method="backfill")

    # Construct altitude values if the flight were to climb / descent throughout
    # group of consecutive nan values. The call to np.minimum / np.maximum cuts
    # the climb / descent off at the terminal altitude of the nan group
    fill_climb = np.minimum(s_ff + nominal_fill, s_bf)
    fill_descent = np.maximum(s_ff - nominal_fill, s_bf)

    # Explicitly determine if the flight is in a climb or descent state
    sign = np.full_like(altitude, np.nan)
    sign[~isna] = np.sign(np.diff(altitude[~isna], append=np.nan))
    sign = pd.Series(sign).fillna(method="ffill")

    # And return the mess
    return np.where(sign == 1, fill_climb, fill_descent)


def _verify_altitude(
    altitude: npt.NDArray[np.float_], nominal_rocd: float, freq: np.timedelta64
) -> None:
    """Confirm that the time derivative of `altitude` does not exceed twice `nominal_rocd`.

    Parameters
    ----------
    altitude : npt.NDArray[np.float_]
        Array of filled altitude values containing nan values.
    nominal_rocd : float
        Nominal rate of climb/descent, in m/s
    freq : np.timedelta64
        Frequency of time index associated to `altitude`.
    """
    dalt = np.diff(altitude)
    dt = freq / np.timedelta64(1, "s")
    rocd = np.abs(dalt / dt)
    if np.any(rocd > 2 * nominal_rocd):
        warnings.warn(
            "Rate of climb/descent values greater than nominal "
            f"({nominal_rocd} m/s) after altitude interpolation"
        )
    if np.any(np.isnan(altitude)):
        warnings.warn(
            f"Found nan values altitude after ({nominal_rocd} m/s) after altitude interpolation"
        )


def filter_altitude(
    altitude: npt.NDArray[np.float_], *, kernel_size: int = 17
) -> npt.NDArray[np.float_]:
    """
    Filter noisy altitude on a single flight.

    Currently runs altitude through a median filter using :func:`scipy.signal.medfilt`
    with ``kernel_size``, then a Savitzky-Golay filter to filter noise.

    .. todo::

        This method assumes that the time interval between altitude points
        (:func:`segment_duration`) is moderately small (e.g. minutes).
        This filter may not work as well when waypoints are close (seconds) or
        farther apart in time (e.g. 30 minutes).

        The optimal altitude filter is a work in a progress
        and may change in the future.

    Parameters
    ----------
    altitude : npt.NDArray[np.float_]
        Altitude signal
    kernel_size : int, optional
        Passed directly to :func:`scipy.signal.medfilt`, by default 11.
        Passed also to :func:`scipy.signal.medfilt`

    Returns
    -------
    npt.NDArray[np.float_]
        Filtered altitude

    Notes
    -----
    Algorithm is derived from :meth:`traffic.core.flight.Flight.filter`.

    The `traffic
    <https://traffic-viz.github.io/api_reference/traffic.core.flight.html#traffic.core.Flight.filter>`_
    algorithm also computes thresholds on sliding windows
    and replaces unacceptable values with NaNs.

    Errors may raised if the ``kernel_size`` is too large.

    See Also
    --------
    :meth:`traffic.core.flight.Flight.filter`
    :func:`scipy.signal.medfilt`
    """  # noqa: E501
    if not len(altitude):
        raise ValueError("Altitude must have non-zero length to filter")

    # The kernel_size must be less than or equal to the number of data points available.
    kernel_size = min(kernel_size, altitude.size)

    # The kernel_size must be odd.
    if (kernel_size % 2) == 0:
        kernel_size -= 1

    # Apply a median filter above a certain threshold
    altitude_filt = scipy.signal.medfilt(altitude, kernel_size=kernel_size)

    # TODO: I think this makes sense because it smooths the climb/descent phases
    altitude_filt = _sg_filter(altitude_filt, window_length=kernel_size)

    # TODO: The right way to do this is with a low pass filter at
    # a reasonable rocd threshold ~300-500 ft/min, e.g.
    # sos = scipy.signal.butter(4, 250, 'low', output='sos')
    # return scipy.signal.sosfilt(sos, altitude_filt1)
    #
    # Remove noise manually
    # only remove above max airport elevation
    d_alt_ft = np.diff(altitude_filt, append=np.nan)
    is_noise = (np.abs(d_alt_ft) <= 25.0) & (altitude_filt > MAX_AIRPORT_ELEVATION)
    altitude_filt[is_noise] = np.round(altitude_filt[is_noise], -3)

    return altitude_filt


def segment_duration(
    time: npt.NDArray[np.datetime64], dtype: np.dtype = np.dtype(np.float32)
) -> npt.NDArray[np.float_]:
    """Calculate the time difference between waypoints.

    ``np.nan`` appended so the length of the output is the same as number of waypoints.

    Parameters
    ----------
    time : npt.NDArray[np.datetime64]
        Waypoint time in ``np.datetime64`` format.
    dtype : np.dtype
        Numpy dtype for time difference.
        Defaults to ``np.float64``

    Returns
    -------
    npt.NDArray[np.float_]
        Time difference between waypoints, [:math:`s`].
        This returns an array with dtype specified by``dtype``.
    """
    out = np.empty_like(time, dtype=dtype)
    out[-1] = np.nan
    out[:-1] = np.diff(time) / np.timedelta64(1, "s")
    return out


def segment_phase(
    rocd: npt.NDArray[np.float_],
    altitude_ft: npt.NDArray[np.float_],
    *,
    threshold_rocd: float = 250.0,
    min_cruise_altitude_ft: float = MIN_CRUISE_ALTITUDE,
) -> npt.NDArray[np.uint8]:
    """Identify the phase of flight (climb, cruise, descent) for each segment.

    Parameters
    ----------
    rocd: pt.NDArray[np.float_]
        Rate of climb and descent across segment, [:math:`ft min^{-1}`].
        See output from :func:`segment_rocd`.
    altitude_ft: npt.NDArray[np.float_]
        Altitude, [:math:`ft`]
    threshold_rocd: float, optional
        ROCD threshold to identify climb and descent, [:math:`ft min^{-1}`].
        Defaults to 250 ft/min.
    min_cruise_altitude_ft: float, optional
        Minimum threshold altitude for cruise, [:math:`ft`]
        This is specific for each aircraft type,
        and can be approximated as 50% of the altitude ceiling.
        Defaults to :attr:`MIN_CRUISE_ALTITUDE`.

    Returns
    -------
    npt.NDArray[np.uint8]
        Array of values enumerating the flight phase.
        See :attr:`flight.FlightPhase` for enumeration.

    Notes
    -----
    Flight data derived from ADS-B and radar sources could contain noise leading
    to small changes in altitude and ROCD. Hence, an arbitrary ``threshold_rocd``
    is specified to identify the different phases of flight.

    The flight phase "level-flight" is when an aircraft is holding at lower altitudes.
    The cruise phase of flight only occurs above a certain threshold altitude.

    See Also
    --------
    :attr:`FlightPhase`
    :func:`segment_rocd`
    """
    nan = np.isnan(rocd)
    cruise = (
        (rocd < threshold_rocd) & (rocd > -threshold_rocd) & (altitude_ft > min_cruise_altitude_ft)
    )
    climb = ~cruise & (rocd > 0)
    descent = ~cruise & (rocd < 0)
    level_flight = ~(nan | cruise | climb | descent)

    phase = np.empty(rocd.shape, dtype=np.uint8)
    phase[cruise] = FlightPhase.CRUISE
    phase[climb] = FlightPhase.CLIMB
    phase[descent] = FlightPhase.DESCENT
    phase[level_flight] = FlightPhase.LEVEL_FLIGHT
    phase[nan] = FlightPhase.NAN

    return phase


def segment_rocd(
    segment_duration: npt.NDArray[np.float_], altitude_ft: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """Calculate the rate of climb and descent (ROCD).

    Parameters
    ----------
    segment_duration: npt.NDArray[np.float_]
        Time difference between waypoints, [:math:`s`].
        Expected to have numeric `dtype`, not `"timedelta64"`.
        See output from :func:`segment_duration`.
    altitude_ft: npt.NDArray[np.float_]
        Altitude of each waypoint, [:math:`ft`]

    Returns
    -------
    npt.NDArray[np.float_]
        Rate of climb and descent over segment, [:math:`ft min^{-1}`]

    See Also
    --------
    :func:`segment_duration`
    """
    dt_min = segment_duration / 60.0

    out = np.empty_like(altitude_ft)
    out[:-1] = np.diff(altitude_ft) / dt_min[:-1]
    out[-1] = np.nan

    return out



"""
Persistent contrail regions (PCR = SAC & ISSR).

Equivalent to (SAC & ISSR)
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, overload

import numpy as np

from pycontrails.core.flight import Flight
from pycontrails.core.met import MetDataArray, MetDataset
from pycontrails.core.met_var import AirTemperature, SpecificHumidity
from pycontrails.core.models import Model
from pycontrails.core.vector import GeoVectorDataset
from pycontrails.models import issr, sac
from pycontrails.physics import thermo
from pycontrails.utils.types import ArrayLike, apply_nan_mask_to_arraylike


@dataclass
class PCRParams(sac.SACParams, issr.ISSRParams):
    """Persistent Contrail Regions (PCR) parameters."""


class PCR(Model):
    """Determine points with likely persistent contrails (PCR).

    Intersection of Ice Super Saturated Regions (ISSR) with regions in which the Schmidt-Appleman
    Criteria (SAC) is satisfied.

    Parameters
    ----------
    met : MetDataset
        Dataset containing "air_temperature", "specific_humidity" variables
    params : dict[str, Any], optional
        Override PCR model parameters with dictionary.
        See :class:`PCRGridParams` for model parameters.
    **params_kwargs
        Override PCR model parameters with keyword arguments.
        See :class:`PCRGridParams` for model parameters.
    """

    name = "pcr"
    long_name = "Persistent contrail regions"
    met_variables = AirTemperature, SpecificHumidity
    default_params = PCRParams

    @overload
    def eval(self, source: Flight, **params: Any) -> Flight:
        ...

    @overload
    def eval(self, source: GeoVectorDataset, **params: Any) -> GeoVectorDataset:
        ...

    @overload
    def eval(self, source: MetDataset | None = None, **params: Any) -> MetDataArray:
        ...

    def eval(
        self, source: GeoVectorDataset | Flight | MetDataset | None = None, **params: Any
    ) -> GeoVectorDataset | Flight | MetDataArray:
        """Evaluate potential contrails regions of the :attr:`met` grid.

        Parameters
        ----------
        source : GeoVectorDataset | Flight | MetDataset | None, optional
            Input GeoVectorDataset or Flight.
            If None, evaluates at the :attr:`met` grid points.

        Returns
        -------
        GeoVectorDataset | Flight | MetDataArray
            Returns 1 in potential contrail regions, 0 everywhere else.
            Returns `np.nan` if interpolating outside meteorology grid.
        **params : Any
            Overwrite model parameters before eval
        Raises
        ------
        NotImplementedError
            Raises if input ``source`` is not supported.
        """

        self.update_params(params)
        self.set_source(source)
        issr_params = {k: v for k, v in self.params.items() if hasattr(issr.ISSR.default_params, k)}
        issr_model = issr.ISSR(self.met, params=issr_params, copy_source=False)
        issr_model.eval(self.source)

        sac_params = {k: v for k, v in self.params.items() if hasattr(sac.SAC.default_params, k)}
        # NOTE: met is not needed here: ISSR already used it
        sac_model = sac.SAC(met=None, params=sac_params, copy_source=False)
        sac_model.eval(self.source)

        pcr_ = _pcr_from_issr_and_sac(self.source.data["issr"], self.source.data["sac"])
        self.source["pcr"] = pcr_
        # Tag output with additional attrs when source is MetDataset
        if isinstance(self.source, MetDataset):
            attrs = {**self.source["issr"].attrs, **self.source["sac"].attrs}
            attrs["description"] = self.long_name
            self.source["pcr"].attrs.update(attrs)
            return self.source["pcr"]

        # In GeoVectorDataset case, return source
        return self.source


def pcr(
    air_temperature: ArrayLike,
    specific_humidity: ArrayLike,
    air_pressure: ArrayLike,
    engine_efficiency: float | ArrayLike,
    ei_h2o: float,
    q_fuel: float,
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    r"""Calculate regions of persistent contrail formation.

    Ice Super Saturated Regions (ISSR) where the Schmidt-Appleman Criteria (SAC) is satisfied.

    Parameters of type :class:`ArrayLike` must have compatible shapes.

    Parameters
    ----------
    air_temperature : ArrayLike
        A sequence or array of temperature values, [:math:`K`]
    specific_humidity : ArrayLike
        A sequence or array of specific humidity values, [:math:`kg_{H_{2}O} \ kg_{air}^{-1}`]
    air_pressure : ArrayLike
        A sequence or array of atmospheric pressure values, [:math:`Pa`].
    engine_efficiency: float | ArrayLike
        Engine efficiency, [:math:`0 - 1`]
    ei_h2o : float
        Emission index of water vapor, [:math:`kg \ kg^{-1}`]
    q_fuel : float
        Specific combustion heat of fuel combustion, [:math:`J \ kg^{-1} \ K^{-1}`]

    Returns
    -------
    pcr : ArrayLike
        PCR state of each point indexed by the :class:`ArrayLike` parameters.
    sac : ArrayLike
        SAC state
    issr : ArrayLike
        ISSR state
    """
    issr_ = issr.issr(air_temperature, specific_humidity, air_pressure)
    G = sac.slope_mixing_line(specific_humidity, air_pressure, engine_efficiency, ei_h2o, q_fuel)
    T_sat_liquid_ = sac.T_sat_liquid(G)
    rh = thermo.rh(specific_humidity, air_temperature, air_pressure)
    rh_crit_sac = sac.rh_critical_sac(air_temperature, T_sat_liquid_, G)
    sac_ = sac.sac(rh, rh_crit_sac)

    pcr_ = _pcr_from_issr_and_sac(issr_, sac_)
    return pcr_, sac_, issr_


def _pcr_from_issr_and_sac(issr_: ArrayLike, sac_: ArrayLike) -> ArrayLike:
    # store nan values to refill after casting
    nan_mask = np.isnan(issr_) | np.isnan(sac_)

    dtype = np.result_type(issr_, sac_)
    pcr_ = ((issr_ > 0.0) & (sac_ > 0.0)).astype(dtype)
    return apply_nan_mask_to_arraylike(pcr_, nan_mask)



"""Contrail Property Calculations."""

from __future__ import annotations

import logging
import warnings
from typing import Any, TypeVar

import numpy as np
import numpy.typing as npt

from pycontrails.models.cocip import radiative_heating
from pycontrails.physics import constants, thermo, units

logger = logging.getLogger(__name__)

####################
# Initial Contrail Properties
####################


def initial_iwc(
    air_temperature: npt.NDArray[np.float_],
    specific_humidity: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    fuel_dist: npt.NDArray[np.float_],
    width: npt.NDArray[np.float_],
    depth: npt.NDArray[np.float_],
    ei_h2o: float,
) -> npt.NDArray[np.float_]:
    r"""
    Estimate the initial contrail ice water content (iwc) before the wake vortex phase.

    Note that the ice water content is replaced by zero if it is negative (dry air),
    and this will end the contrail life-cycle in subsequent steps.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    specific_humidity : npt.NDArray[np.float_]
        ambient specific humidity for each waypoint, [:math:`kg_{H_{2}O}/kg_{air}`]
    air_pressure : npt.NDArray[np.float_]
        initial pressure altitude at each waypoint, before the wake vortex phase, [:math:`Pa`]
    fuel_dist : npt.NDArray[np.float_]
        fuel consumption of the flight segment per distance traveled, [:math:`kg m^{-1}`]
    width : npt.NDArray[np.float_]
        initial contrail width, [:math:`m`]
    depth : npt.NDArray[np.float_]
        initial contrail depth, [:math:`m`]
    ei_h2o : float
        water vapor emissions index of fuel, [:math:`kg_{H_{2}O} \ kg_{fuel}^{-1}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Initial contrail ice water content (iwc) at the original waypoint
        before the wake vortex phase, [:math:`kg_{H_{2}O}/kg_{air}`].
        Returns zero if iwc is is negative (dry air).
    """
    q_sat = thermo.q_sat_ice(air_temperature, air_pressure)
    q_exhaust_ = q_exhaust(air_temperature, air_pressure, fuel_dist, width, depth, ei_h2o)
    return np.maximum(q_exhaust_ + specific_humidity - q_sat, 0.0)


def q_exhaust(
    air_temperature: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    fuel_dist: npt.NDArray[np.float_],
    width: npt.NDArray[np.float_],
    depth: npt.NDArray[np.float_],
    ei_h2o: float,
) -> npt.NDArray[np.float_]:
    r"""
    Calculate the specific humidity released by water vapor from aircraft emissions.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    air_pressure : npt.NDArray[np.float_]
        initial pressure altitude at each waypoint, before the wake vortex phase, [:math:`Pa`]
    fuel_dist : npt.NDArray[np.float_]
        fuel consumption of the flight segment per distance travelled, [:math:`kg m^{-1}`]
    width : npt.NDArray[np.float_]
        initial contrail width, [:math:`m`]
    depth : npt.NDArray[np.float_]
        initial contrail depth, [:math:`m`]
    ei_h2o : float
        water vapor emissions index of fuel, [:math:`kg_{H_{2}O} \ kg_{fuel}^{-1}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Humidity released by water vapour from aircraft emissions, [:math:`kg_{H_{2}O}/kg_{air}`]
    """
    rho_air = thermo.rho_d(air_temperature, air_pressure)
    return (ei_h2o * fuel_dist) / ((np.pi / 4) * width * depth * rho_air)


def iwc_adiabatic_heating(
    air_temperature: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_pressure_1: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the change in ice water content due to adiabatic heating from the wake vortex phase.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    air_pressure : npt.NDArray[np.float_]
        initial pressure altitude at each waypoint, before the wake vortex phase, [:math:`Pa`]
    air_pressure_1 : npt.NDArray[np.float_]
        pressure altitude at each waypoint, after the wake vortex phase, [:math:`Pa`]

    Returns
    -------
    npt.NDArray[np.float_]
        Change in ice water content due to adiabatic heating from the wake
        vortex phase, [:math:`kg_{H_{2}O}/kg_{air}`]
    """
    p_ice = thermo.e_sat_ice(air_temperature)
    air_temperature_1 = temperature_adiabatic_heating(air_temperature, air_pressure, air_pressure_1)
    p_ice_1 = thermo.e_sat_ice(air_temperature_1)
    return np.maximum(
        (constants.R_d / constants.R_v) * ((p_ice_1 / air_pressure_1) - (p_ice / air_pressure)), 0
    )


def temperature_adiabatic_heating(
    air_temperature: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_pressure_1: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """Calculate the ambient air temperature for each waypoint after the wake vortex phase.

    This calculation accounts for adiabatic heating.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    air_pressure : npt.NDArray[np.float_]
        initial pressure altitude at each waypoint, before the wake vortex phase, [:math:`Pa`]
    air_pressure_1 : npt.NDArray[np.float_]
        pressure altitude at each waypoint, after the wake vortex phase, [:math:`Pa`]

    Returns
    -------
    npt.NDArray[np.float_]
        ambient air temperature after the wake vortex phase, [:math:`K`]

    Notes
    -----
    Level 1, see Figure 1 of :cite:`schumannContrailCirrusPrediction2012`

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    exponent = (constants.gamma - 1) / constants.gamma
    return air_temperature * (air_pressure_1 / air_pressure) ** exponent


def iwc_post_wake_vortex(
    iwc: npt.NDArray[np.float_], iwc_ad: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the ice water content after the wake vortex phase (``iwc_1``).

    ``iwc_1`` is calculated by subtracting the initial iwc before the wake vortex phase (``iwc``)
    by the change in iwc from adiabatic heating experienced during the wake vortex phase.

    Note that the iwc is replaced by zero if it is negative (dry air),
    and this will end the contrail lifecycle in subsequent steps.

    Parameters
    ----------
    iwc : npt.NDArray[np.float_]
        initial ice water content at each waypoint before the wake vortex
        phase, [:math:`kg_{H_{2}O}/kg_{air}`]
    iwc_ad : npt.NDArray[np.float_]
        change in iwc from adiabatic heating during the wake vortex
        phase, [:math:`kg_{H_{2}O}/kg_{air}`]

    Returns
    -------
    npt.NDArray[np.float_]
        ice water content after the wake vortex phase, ``iwc_1``, [:math:`kg_{H_{2}O}/kg_{air}`]

    Notes
    -----
    Level 1, see Figure 1 of :cite:`schumannContrailCirrusPrediction2012`

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """
    return np.maximum((iwc - iwc_ad), 0)


def ice_particle_number(
    nvpm_ei_n: npt.NDArray[np.float_],
    fuel_dist: npt.NDArray[np.float_],
    iwc: npt.NDArray[np.float_],
    iwc_1: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    T_crit_sac: npt.NDArray[np.float_],
    min_ice_particle_number_nvpm_ei_n: float,
) -> npt.NDArray[np.float_]:
    """Calculate the initial number of ice particles per distance after the wake vortex phase.

    The initial number of ice particle per distance is calculated from the black
    carbon number emissions index ``nvpm_ei_n`` and fuel burn per distance ``fuel_dist``.
    Note that a lower bound for ``nvpm_ei_n`` is set at ``1e13`` :math:`kg^{-1}` to account
    for the activation of ambient aerosol particles and organic volatile particles.

    Parameters
    ----------
    nvpm_ei_n : npt.NDArray[np.float_]
        black carbon number emissions index, [:math:`kg^{-1}`]
    fuel_dist : npt.NDArray[np.float_]
        fuel consumption of the flight segment per distance traveled, [:math:`kg m^{-1}`]
    iwc : npt.NDArray[np.float_]
        initial ice water content at each flight waypoint before the wake vortex
        phase, [:math:`kg_{H_{2}O}/kg_{air}`]
    iwc_1 : npt.NDArray[np.float_]
        ice water content after the wake vortex phase, [:math:`kg_{H_{2}O}/kg_{air}`]
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    T_crit_sac : npt.NDArray[np.float_]
        estimated Schmidt-Appleman temperature threshold for contrail formation, [:math:`K`]
    min_ice_particle_number_nvpm_ei_n : float
        lower bound for nvpm_ei_n to account for ambient aerosol particles for
        newer engines [:math:`kg^{-1}`]

    Returns
    -------
    npt.NDArray[np.float_]
        initial number of ice particles per distance after the wake vortex phase, [:math:`# m^{-1}`]
    """
    f_surv = ice_particle_survival_factor(iwc, iwc_1)
    f_activation = ice_particle_activation_rate(air_temperature, T_crit_sac)
    nvpm_ei_n_activated = nvpm_ei_n * f_activation
    return fuel_dist * np.maximum(nvpm_ei_n_activated, min_ice_particle_number_nvpm_ei_n) * f_surv


def ice_particle_activation_rate(
    air_temperature: npt.NDArray[np.float_], T_crit_sac: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the activation rate of black carbon particles to contrail ice crystals.

    The activation rate is calculated as a function of the difference between
    the ambient temperature and the Schmidt-Appleman threshold temperature ``T_crit_sac``.

    Parameters
    ----------
    air_temperature : npt.NDArray[np.float_]
        ambient temperature at each waypoint before wake_wortex, [:math:`K`]
    T_crit_sac : npt.NDArray[np.float_]
        estimated Schmidt-Appleman temperature threshold for contrail formation, [:math:`K`]

    Returns
    -------
    npt.NDArray[np.float_]
        Proportion of black carbon particles that activates to contrail ice parties.

    Notes
    -----
    The equation is not published but based on the raw data
    from :cite:`brauerAirborneMeasurementsContrail2021`.

    References
    ----------
    - :cite:`brauerAirborneMeasurementsContrail2021`
    """
    d_temp = air_temperature - T_crit_sac
    d_temp = np.minimum(d_temp, 0)

    # NOTE: It seems somewhat unnecessary to do this additional "rounding down"
    # of d_temp for values below -5. As d_temp near -5, the activation rate approaches
    # 1. This additional rounding injects a small jump discontinuity into the activation rate that
    # likely does not match reality. I suggest removing the line below. This will change
    # model outputs roughly at 0.001 - 0.1%.
    d_temp[d_temp < -5] = -np.inf
    return -0.661 * np.exp(d_temp) + 1


def ice_particle_survival_factor(
    iwc: npt.NDArray[np.float_], iwc_1: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Estimate the fraction of contrail ice particle number that survive the wake vortex phase.

    CoCiP assumes that this fraction is proportional to the change in ice water content
    (``iwc_1 - iwc``) before and after the wake vortex phase.

    Parameters
    ----------
    iwc : npt.NDArray[np.float_]
        initial ice water content at each waypoint before the wake vortex
        phase, [:math:`kg_{H_{2}O}/kg_{air}`]
    iwc_1 : npt.NDArray[np.float_]
        ice water content after the wake vortex phase, [:math:`kg_{H_{2}O}/kg_{air}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Fraction of contrail ice particle number that survive the wake vortex phase.
    """
    f_surv = np.empty_like(iwc)
    is_positive = (iwc > 0) & (iwc_1 > 0)

    f_surv[is_positive] = iwc_1[is_positive] / iwc[is_positive]
    f_surv[is_positive] = np.minimum(f_surv[is_positive], 1)
    f_surv[~is_positive] = 0.5

    return f_surv


def initial_persistent(
    iwc_1: npt.NDArray[np.float_], rhi_1: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Determine if waypoints have persistent contrails.

    Conditions for persistent initial_contrails:

    1. ice water content at level 1: ``1e-12 < iwc < 1e10``
    2. rhi at level 1: ``0 < rhi < 1e10``

    .. versionchanged:: 0.25.1
        Returned array now has floating dtype. This is consistent with other filtering
        steps in the CoCiP model (ie, ``sac``).

    Parameters
    ----------
    iwc_1 : npt.NDArray[np.float_]
        ice water content after the wake vortex phase, [:math:`kg_{H_{2}O}/kg_{air}`]
    rhi_1 : npt.NDArray[np.float_]
        relative humidity with respect to ice after the wake vortex phase

    Returns
    -------
    npt.NDArray[np.float_]
        Mask of waypoints with persistent contrails. Waypoints with persistent contrails
        will have value 1.

    Notes
    -----
    The RHi at level 1 does not need to be above 100% if the iwc > 0 kg/kg. If iwc > 0
    and RHi < 100%, the contrail lifetime will not end immediately as the ice particles
    will gradually evaporate with the rate depending on the background RHi.
    """
    out = (iwc_1 > 1e-12) & (iwc_1 < 1e10) & (rhi_1 > 0) & (rhi_1 < 1e10)
    dtype = np.result_type(iwc_1, rhi_1)
    return out.astype(dtype)


def contrail_persistent(
    latitude: npt.NDArray[np.float_],
    altitude: npt.NDArray[np.float_],
    segment_length: npt.NDArray[np.float_],
    age: npt.NDArray[np.timedelta64],
    tau_contrail: npt.NDArray[np.float_],
    n_ice_per_m3: npt.NDArray[np.float_],
    params: dict[str, Any],
) -> npt.NDArray[np.bool_]:
    r"""
    Determine surviving contrail segments after time integration step.

    A contrail waypoint reaches its end of life if any of the following conditions hold:

    1. Contrail age exceeds ``max_age``
    2. Contrail optical depth lies outside of interval ``[min_tau, max_tau]``
    3. Ice number density lies outside of interval ``[min_n_ice_per_m3, max_n_ice_per_m3]``
    4. Altitude lies outside of the interval ``[min_altitude_m, max_altitude_m]``
    5. Segment length exceeds ``max_seg_length_m``
    6. Latitude values are within 1 degree of the north or south pole

    This function warns if all values in the ``tau_contrail`` array are nan.

    .. versionchanged:: 0.25.10

        Extreme values of ``latitude`` (ie, close to the north or south pole) now
        create an end of life condition. This check helps address issues related to
        divergence in the polar regions. (With large enough integration time delta,
        it is still possible for post-advection latitude values to lie outside of
        [-90, 90], but this is no longer possible with typical parameters and wind
        speeds.)

    Parameters
    ----------
    latitude : npt.NDArray[np.float_]
        Contrail latitude, [:math:`\deg`]
    altitude : npt.NDArray[np.float_]
        Contrail altitude, [:math:`m`]
    segment_length : npt.NDArray[np.float_]
        Contrail segment length, [:math:`m`]
    age : npt.NDArray[np.timedelta64]
        Contrail age
    tau_contrail : npt.NDArray[np.float_]
        Contrail optical depth
    n_ice_per_m3 : npt.NDArray[np.float_]
        Contrail ice particle number per volume of air, [:math:`# m^{-3}`]
    params : dict[str, Any]
        Dictionary of :class:`CocipParams` parameters determining the
        conditions for end of contrail life.

    Returns
    -------
    npt.NDArray[np.bool_]
        Boolean array indicating surviving contrails. Persisting contrails
        will be marked as True.
    """
    status_1 = _within_range(age, max=params["max_age"])
    status_2 = _within_range(tau_contrail, max=params["max_tau"], min=params["min_tau"])
    status_3 = _within_range(
        n_ice_per_m3, max=params["max_n_ice_per_m3"], min=params["min_n_ice_per_m3"]
    )
    status_4 = _within_range(altitude, max=params["max_altitude_m"], min=params["min_altitude_m"])
    status_5 = _within_range(segment_length, max=params["max_seg_length_m"])
    status_6 = _within_range(latitude, max=89.0, min=-89.0)  # type: ignore[type-var]

    logger.debug(
        "Survival stats. age: %s, tau: %s, ice: %s, altitude: %s, segment: %s, latitude: %s",
        np.sum(status_1),
        np.sum(status_2),
        np.sum(status_3),
        np.sum(status_4),
        np.sum(status_5),
        np.sum(status_6),
    )
    tau_contrail_nan = np.isnan(tau_contrail)
    logger.debug(
        "Fraction of nan in tau_contrail: %s / %s",
        tau_contrail_nan.sum(),
        tau_contrail_nan.size,
    )
    if np.all(tau_contrail_nan):
        warnings.warn(
            "All tau_contrail values are nan. This may be due to waypoints "
            "all lying outside of the met interpolation grid. It could "
            "indicate an issue with interpolation, or an insufficient "
            "met domain."
        )
    return status_1 & status_2 & status_3 & status_4 & status_5 & status_6


T = TypeVar("T", np.float_, np.timedelta64)


def _within_range(
    val: npt.NDArray[T],
    max: npt.NDArray[T] | T | None = None,
    min: npt.NDArray[T] | T | None = None,
) -> npt.NDArray[np.bool_]:
    """
    Check if the input values (val) are each within the specified range.

    If both ``max`` and ``min`` are None, a literal constant True is returned.

    Parameters
    ----------
    val : np.ndarray
        value of selected contrail property
    max : np.ndarray | float | np.timedelta64 | None, optional
        Upper bound. If None, no upper bound is imposed. None by default.
    min : np.ndarray | float | np.timedelta64 | None, optional
        Lower bound. If None, no lower bound is imposed. None by default.

    Returns
    -------
    npt.NDArray[np.bool_]
        Mask of waypoints. Waypoints with values within the specified range will be marked as true.
    """
    cond: npt.NDArray[np.bool_] = True  # type: ignore[assignment]
    if min is not None:
        cond &= val >= min
    if max is not None:
        cond &= val <= max
    return cond


####################
# Contrail Properties
####################


def contrail_edges(
    lon: npt.NDArray[np.float_],
    lat: npt.NDArray[np.float_],
    sin_a: npt.NDArray[np.float_],
    cos_a: npt.NDArray[np.float_],
    width: npt.NDArray[np.float_],
) -> tuple[
    npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]
]:
    """
    Calculate the longitude and latitude of the contrail edges to account for contrail spreading.

    (lon_edge_l, lat_edge_l)        x---------------------

    (Contrail midpoint: lon, lat)   X===================== ->

    (lon_edge_r, lat_edge_r)        x---------------------


    Parameters
    ----------
    lon : npt.NDArray[np.float_]
        longitude of contrail waypoint, degrees
    lat : npt.NDArray[np.float_]
        latitude of contrail waypoint, degrees
    sin_a : npt.NDArray[np.float_]
        sin(a), where a is the angle between the plume and the longitudinal axis
    cos_a : npt.NDArray[np.float_]
        cos(a), where a is the angle between the plume and the longitudinal axis
    width : npt.NDArray[np.float_]
        contrail width at each waypoint, [:math:`m`]

    Returns
    -------
    tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]]
        (lon_edge_l, lat_edge_l, lon_edge_r, lat_edge_r), longitudes and latitudes
        at the left and right edges of the contrail, degrees
    """  # noqa: E501
    dlon = units.m_to_longitude_distance(width * sin_a * 0.5, lat)
    dlat = units.m_to_latitude_distance(width * cos_a * 0.5)

    lon_edge_l = lon - dlon
    lat_edge_l = lat + dlat
    lon_edge_r = lon + dlon
    lat_edge_r = lat - dlat

    return (lon_edge_l, lat_edge_l, lon_edge_r, lat_edge_r)


def contrail_vertices(
    lon: npt.NDArray[np.float_],
    lat: npt.NDArray[np.float_],
    sin_a: npt.NDArray[np.float_],
    cos_a: npt.NDArray[np.float_],
    width: npt.NDArray[np.float_],
    segment_length: npt.NDArray[np.float_],
) -> tuple[
    npt.NDArray[np.float_],
    npt.NDArray[np.float_],
    npt.NDArray[np.float_],
    npt.NDArray[np.float_],
    npt.NDArray[np.float_],
    npt.NDArray[np.float_],
    npt.NDArray[np.float_],
    npt.NDArray[np.float_],
]:
    """
    Calculate the longitude and latitude of the contrail vertices.

    This is equivalent to running :meth:`contrail_edges` at each contrail waypoint
    and associating the next continuous waypoint with the previous.
    This method is helpful when you want to treat each contrail waypoint independently.

    (lon_1, lat_1)                  x--------------------x   (lon_4, lat_4)

    (Contrail waypoint: lon, lat)   X==================== ->

    (lon_2, lat_2)                  x--------------------x   (lon_3, lat_3)


    Parameters
    ----------
    lon : npt.NDArray[np.float_]
        longitude of contrail waypoint, degrees
    lat : npt.NDArray[np.float_]
        latitude of contrail waypoint, degrees
    sin_a : npt.NDArray[np.float_]
        sin(a), where a is the angle between the plume and the longitudinal axis
    cos_a : npt.NDArray[np.float_]
        cos(a), where a is the angle between the plume and the longitudinal axis
    width : npt.NDArray[np.float_]
        contrail width at each waypoint, [:math:`m`]
    segment_length : npt.NDArray[np.float_]
        contrail length at each waypoint, [:math:`m`]

    Returns
    -------
    tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]]
        (lon_1, lat_1, lon_2, lat_2, lon_3, lat_3, lon_4, lat_4) degrees
    """  # noqa: E501
    dlon_width = units.m_to_longitude_distance(width * sin_a * 0.5, lat)
    dlat_width = units.m_to_latitude_distance(width * cos_a * 0.5)

    # using "lat" as mean here is a little inaccurate, but its a close approx
    dlon_length = units.m_to_longitude_distance(segment_length * cos_a, lat)
    dlat_length = units.m_to_latitude_distance(segment_length * sin_a)

    lon_1 = lon - dlon_width
    lon_2 = lon + dlon_width
    lon_3 = lon + dlon_width + dlon_length
    lon_4 = lon - dlon_width + dlon_length

    lat_1 = lat + dlat_width
    lat_2 = lat - dlat_width
    lat_3 = lat - dlat_width + dlat_length
    lat_4 = lat + dlat_width + dlat_length

    return (lon_1, lat_1, lon_2, lat_2, lon_3, lat_3, lon_4, lat_4)


def plume_effective_cross_sectional_area(
    width: npt.NDArray[np.float_], depth: npt.NDArray[np.float_], sigma_yz: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the effective cross-sectional area of the contrail plume (``area_eff``).

    ``sigma_yy``, ``sigma_zz`` and ``sigma_yz`` are the parameters governing the
    contrail plume's temporal evolution.

    Parameters
    ----------
    width : npt.NDArray[np.float_]
        contrail width at each waypoint, [:math:`m`]
    depth : npt.NDArray[np.float_]
        contrail depth at each waypoint, [:math:`m`]
    sigma_yz : npt.NDArray[np.float_]
        temporal evolution of the contrail plume parameters

    Returns
    -------
    npt.NDArray[np.float_]
        effective cross-sectional area of the contrail plume, [:math:`m^{2}`]
    """
    sigma_yy = 0.125 * (width**2)
    sigma_zz = 0.125 * (depth**2)
    return new_effective_area_from_sigma(sigma_yy, sigma_zz, sigma_yz)


def plume_effective_depth(
    width: npt.NDArray[np.float_], area_eff: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the effective depth of the contrail plume (``depth_eff``).

    ``depth_eff`` is calculated from the effective cross-sectional area (``area_eff``)
    and the contrail width.

    Parameters
    ----------
    width : npt.NDArray[np.float_]
        contrail width at each waypoint, [:math:`m`]
    area_eff : npt.NDArray[np.float_]
        effective cross-sectional area of the contrail plume, [:math:`m^{2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        effective depth of the contrail plume, [:math:`m`]
    """
    return area_eff / width


def plume_mass_per_distance(
    area_eff: npt.NDArray[np.float_], rho_air: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the contrail plume mass per unit length.

    Parameters
    ----------
    area_eff : npt.NDArray[np.float_]
        effective cross-sectional area of the contrail plume, [:math:`m^{2}`]
    rho_air : npt.NDArray[np.float_]
        density of air for each waypoint, [:math:`kg m^{-3}`]

    Returns
    -------
    npt.NDArray[np.float_]
        contrail plume mass per unit length, [:math:`kg m^{-1}`]
    """
    return area_eff * rho_air


def ice_particle_number_per_volume_of_plume(
    n_ice_per_m: npt.NDArray[np.float_], area_eff: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the number of contrail ice particles per volume of plume (``n_ice_per_vol``).

    Parameters
    ----------
    n_ice_per_m : npt.NDArray[np.float_]
        number of ice particles per distance at time t, [:math:`m^{-1}`]
    area_eff : npt.NDArray[np.float_]
        effective cross-sectional area of the contrail plume, [:math:`m^{2}`]

    Returns
    -------
    npt.NDArray[np.float_]
        number of ice particles per volume of contrail plume at time t, [:math:`# m^{-3}`]
    """
    return n_ice_per_m / area_eff


def ice_particle_number_per_mass_of_air(
    n_ice_per_vol: npt.NDArray[np.float_], rho_air: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the number of contrail ice particles per mass of air.

    Parameters
    ----------
    n_ice_per_vol : npt.NDArray[np.float_]
        number of ice particles per volume of contrail plume at time t, [:math:`# m^{-3}`]
    rho_air : npt.NDArray[np.float_]
        density of air for each waypoint, [:math:`kg m^{-3}`]

    Returns
    -------
    npt.NDArray[np.float_]
        number of ice particles per mass of air at time t, [:math:`# kg^{-1}`]
    """
    return n_ice_per_vol / rho_air


def ice_particle_volume_mean_radius(
    iwc: npt.NDArray[np.float_], n_ice_per_kg_air: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate the ice particle volume mean radius.

    Parameters
    ----------
    iwc : npt.NDArray[np.float_]
        contrail ice water content, i.e., contrail ice mass per
        kg of air, [:math:`kg_{H_{2}O}/kg_{air}`]

    n_ice_per_kg_air : npt.NDArray[np.float_]
        number of ice particles per mass of air, [:math:`# kg^{-1}`]

    Returns
    -------
    npt.NDArray[np.float_]
        ice particle volume mean radius, [:math:`m`]

    Notes
    -----
    ``r_ice_vol`` is the mean radius of a sphere that has the same volume as the
    contrail ice particle.

    ``r_ice_vol`` calculated by dividing the total volume of contrail
    ice particle per kg of air (``total_ice_volume``, :math:`m**3/kg-air`) with the
    number of contrail ice particles per kg of air (``n_ice_per_kg_air``, :math:`#/kg-air`).
    """
    total_ice_volume = iwc / constants.rho_ice
    r_ice_vol = ((3 / (4 * np.pi)) * (total_ice_volume / n_ice_per_kg_air)) ** (1 / 3)
    zero_negative_values = iwc <= 0
    r_ice_vol[zero_negative_values] = iwc[zero_negative_values]
    return np.maximum(r_ice_vol, 1e-10)


def ice_particle_terminal_fall_speed(
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    r_ice_vol: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the terminal fall speed of contrail ice particles.

    ``v_t`` is calculated based on a parametric model
    from :cite:`spichtingerModellingCirrusClouds2009`, using inputs of pressure
    level, ambient temperature and the ice particle volume mean radius.

    Parameters
    ----------
    air_pressure : npt.NDArray[np.float_]
        Pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature for each waypoint, [:math:`K`]
    r_ice_vol : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Terminal fall speed of contrail ice particles, [:math:`m s^{-1}`]

    References
    ----------
    - :cite:`spichtingerModellingCirrusClouds2009`
    """
    ipm = ice_particle_mass(r_ice_vol)

    alpha = np.full_like(r_ice_vol, np.nan)

    # For ice particle mass >= 4.264e-8 kg
    particle_mass = ipm >= 4.264e-8
    alpha[particle_mass] = 8.80 * ipm[particle_mass] ** 0.096

    # For ice particle mass in [2.166e-9 kg, 4.264e-8 kg)
    particle_mass = (ipm < 4.264e-8) & (ipm >= 2.166e-9)
    alpha[particle_mass] = 329.8 * ipm[particle_mass] ** 0.31

    # For ice particle mass in [2.146e-13 kg, 2.166e-9 kg)
    particle_mass = (ipm < 2.166e-9) & (ipm >= 2.146e-13)
    alpha[particle_mass] = 63292.4 * ipm[particle_mass] ** 0.57

    # For ice particle mass < 2.146e-13 kg
    particle_mass = ipm < 2.146e-13
    alpha[particle_mass] = 735.4 * ipm[particle_mass] ** 0.42

    return alpha * (30000 / air_pressure) ** 0.178 * (233 / air_temperature) ** 0.394


def ice_particle_mass(r_ice_vol: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """
    Calculate the contrail ice particle mass.

    It is calculated by multiplying the mean ice particle volume with the density of ice

    Parameters
    ----------
    r_ice_vol : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Mean contrail ice particle mass, [:math:`kg`]
    """
    return ((4 / 3) * np.pi * r_ice_vol**3) * constants.rho_ice


def horizontal_diffusivity(
    ds_dz: npt.NDArray[np.float_], depth: npt.NDArray[np.float_]
) -> npt.NDArray[np.float_]:
    """
    Calculate contrail horizontal diffusivity.

    Parameters
    ----------
    ds_dz : npt.NDArray[np.float_]
        Total wind shear (eastward and northward winds) with respect
        to altitude (``dz``), [:math:`m s^{-1} / Pa`]
    depth : npt.NDArray[np.float_]
        Contrail depth at each waypoint, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        horizontal diffusivity, [:math:`m^{2} s^{-1}`]

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`

    Notes
    -----
    Accounts for the turbulence-induced diffusive contrail spreading in
    the horizontal direction.
    """
    return 0.1 * ds_dz * depth**2


def vertical_diffusivity(
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    dT_dz: npt.NDArray[np.float_],
    depth_eff: npt.NDArray[np.float_],
    terminal_fall_speed: npt.NDArray[np.float_],
    sedimentation_impact_factor: npt.NDArray[np.float_] | float,
    eff_heat_rate: npt.NDArray[np.float_] | None,
) -> npt.NDArray[np.float_]:
    """
    Calculate contrail vertical diffusivity.

    Parameters
    ----------
    air_pressure : npt.NDArray[np.float_]
        Pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature : npt.NDArray[np.float_]
        Ambient temperature for each waypoint, [:math:`K`]
    dT_dz : npt.NDArray[np.float_]
        Temperature gradient with respect to altitude (dz), [:math:`K m^{-1}`]
    depth_eff : npt.NDArray[np.float_]
        Effective depth of the contrail plume, [:math:`m`]
    terminal_fall_speed : npt.NDArray[np.float_]
        Terminal fall speed of contrail ice particles, [:math:`m s^{-1}`]
    sedimentation_impact_factor : float
        Enhancement parameter denoted by `f_T` in eq. (35) Schumann (2012).
    eff_heat_rate: npt.NDArray[np.float_] | None
        Effective heating rate, i.e., rate of which the contrail plume
        is heated, [:math:`K s^{-1}`]. If None is passed, the radiative
        heating effects on contrail cirrus properties are not included.

    Returns
    -------
    npt.NDArray[np.float_]
        vertical diffusivity, [:math:`m^{2} s^{-1}`]

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    - :cite:`schumannAviationinducedCirrusRadiation2013`

    Notes
    -----
    Accounts for the turbulence-induced diffusive contrail spreading in the vertical direction.
    See eq. (35) of :cite:`schumannContrailCirrusPrediction2012`.

    The first term in Eq. (35) of Schumann (2012)
    (c_V * w'_N^2 / N_BV, where c_V = 0.2 and w'_N^2 = 0.1) is different
    than outlined below. Here, a constant of 0.01 is used when radiative
    heating effects are not activated. This update comes from Schumann and
    Graf (2013), which found that the original formulation estimated thinner
    contrails relative to satellite observations. The vertical diffusivity
    was enlarged so that the simulated contrails are more consistent with observations.
    """
    n_bv = thermo.brunt_vaisala_frequency(air_pressure, air_temperature, dT_dz)
    n_bv.clip(min=0.001, out=n_bv)

    cvs: npt.NDArray[np.float_] | float
    if eff_heat_rate is not None:
        cvs = radiative_heating.convective_velocity_scale(depth_eff, eff_heat_rate, air_temperature)
        cvs.clip(min=0.01, out=cvs)
    else:
        cvs = 0.01

    return cvs / n_bv + sedimentation_impact_factor * terminal_fall_speed * depth_eff


####################
# Ice particle losses
####################


def particle_losses_aggregation(
    r_ice_vol: npt.NDArray[np.float_],
    terminal_fall_speed: npt.NDArray[np.float_],
    area_eff: npt.NDArray[np.float_],
    agg_efficiency: float = 1.0,
) -> npt.NDArray[np.float_]:
    """
    Calculate the rate of contrail ice particle losses due to sedimentation-induced aggregation.

    Parameters
    ----------
    r_ice_vol : npt.NDArray[np.float_]
        Ice particle volume mean radius, [:math:`m`]
    terminal_fall_speed : npt.NDArray[np.float_]
        Terminal fall speed of contrail ice particles, [:math:`m s^{-1}`]
    area_eff : npt.NDArray[np.float_]
        Effective cross-sectional area of the contrail plume, [:math:`m^{2}`]
    agg_efficiency : float, optional
        Aggregation efficiency

    Returns
    -------
    npt.NDArray[np.float_]
        Rate of contrail ice particle losses due to sedimentation-induced
        aggregation, [:math:`# s^{-1}`]

    Notes
    -----
    The aggregation efficiency (``agg_efficiency = 1``) was calibrated based on
    the observed lifetime and optical properties from the Contrail Library (COLI)
    database (:cite:`schumannPropertiesIndividualContrails2017`).

    References
    ----------
    - :cite:`schumannPropertiesIndividualContrails2017`
    """
    return (8 * agg_efficiency * np.pi * r_ice_vol**2 * terminal_fall_speed) / area_eff


def particle_losses_turbulence(
    width: npt.NDArray[np.float_],
    depth: npt.NDArray[np.float_],
    depth_eff: npt.NDArray[np.float_],
    diffuse_h: npt.NDArray[np.float_],
    diffuse_v: npt.NDArray[np.float_],
    turb_efficiency: float = 0.1,
) -> npt.NDArray[np.float_]:
    """
    Calculate the rate of contrail ice particle losses due to plume-internal turbulence.

    Parameters
    ----------
    width : npt.NDArray[np.float_]
        Contrail width at each waypoint, [:math:`m`]
    depth : npt.NDArray[np.float_]
        Contrail depth at each waypoint, [:math:`m`]
    depth_eff : npt.NDArray[np.float_]
        Effective depth of the contrail plume, [:math:`m`]
    diffuse_h : npt.NDArray[np.float_]
        Horizontal diffusivity, [:math:`m^{2} s^{-1}`]
    diffuse_v : npt.NDArray[np.float_]
        Vertical diffusivity, [:math:`m^{2} s^{-1}`]
    turb_efficiency : float, optional
        Turbulence sublimation efficiency

    Returns
    -------
    npt.NDArray[np.float_]
        Rate of contrail ice particle losses due to plume-internal turbulence, [:math:`# s^{-1}`]

    Notes
    -----
    The turbulence sublimation efficiency (``turb_efficiency = 0.1``) was calibrated
    based on the observed lifetime and optical properties from the Contrail Library (COLI)
    database (:cite:`schumannPropertiesIndividualContrails2017`).

    References
    ----------
    - :cite:`schumannPropertiesIndividualContrails2017`
    """
    inner_term = (diffuse_h / (np.maximum(width, depth)) ** 2) + (diffuse_v / depth_eff**2)
    return turb_efficiency * np.abs(inner_term)


####################
# Optical properties
####################


def contrail_optical_depth(
    r_ice_vol: npt.NDArray[np.float_],
    n_ice_per_m: npt.NDArray[np.float_],
    width: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the contrail optical depth for each waypoint.

    Parameters
    ----------
    r_ice_vol : npt.NDArray[np.float_]
        ice particle volume mean radius, [:math:`m`]
    n_ice_per_m : npt.NDArray[np.float_]
        Number of contrail ice particles per distance, [:math:`m^{-1}`]
    width : npt.NDArray[np.float_]
        Contrail width, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Contrail optical depth
    """
    q_ext = scattering_extinction_efficiency(r_ice_vol)
    tau_contrail = constants.c_r * np.pi * r_ice_vol**2 * (n_ice_per_m / width) * q_ext

    bool_small = r_ice_vol <= 1e-9
    tau_contrail[bool_small] = 0
    return np.maximum(tau_contrail, 0)


def scattering_extinction_efficiency(r_ice_vol: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """
    Calculate the scattering extinction efficiency (``q_ext``) based on Mie-theory.

    Parameters
    ----------
    r_ice_vol : npt.NDArray[np.float_]
        ice particle volume mean radius, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        scattering extinction efficiency

    References
    ----------
    - https://en.wikipedia.org/wiki/Mie_scattering
    """
    phase_delay = light_wave_phase_delay(r_ice_vol)
    return 2 - (4 / phase_delay) * (np.sin(phase_delay) - ((1 - np.cos(phase_delay)) / phase_delay))


def light_wave_phase_delay(r_ice_vol: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
    """
    Calculate the phase delay of the light wave passing through the contrail ice particle.

    Parameters
    ----------
    r_ice_vol : npt.NDArray[np.float_]
        ice particle volume mean radius, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        phase delay of the light wave passing through the contrail ice particle

    References
    ----------
    - https://en.wikipedia.org/wiki/Mie_scattering
    """
    phase_delay = (4 * np.pi * (constants.mu_ice - 1) / constants.lambda_light) * r_ice_vol
    return np.minimum(phase_delay, 100)


#######################################################
# Contrail evolution: second-order Runge Kutta scheme
#######################################################
# Notation "t1" implies properties at the start of the time step (before the time integration step)
# Notation "t2" implies properties at the end of the time step (after the time integration step)


def segment_length_ratio(
    seg_length_t1: npt.NDArray[np.float_],
    seg_length_t2: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """Calculate the ratio of contrail segment length pre-advection to post-advection.

    Parameters
    ----------
    seg_length_t1 : npt.NDArray[np.float_]
        Segment length of contrail waypoint at the start of the time step, [:math:`m`]
    seg_length_t2 : npt.NDArray[np.float_]
        Segment length of contrail waypoint after time step and advection, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Ratio of segment length before advection to segment length after advection.

    Notes
    -----
    This implementation differs from the original fortran implementation.
    Instead of taking a geometric mean between
    the previous and following segments, a simple ratio is computed.

    For terminal waypoints along a flight trajectory, the associated segment length is 0. In this
    case, the segment ratio is set to 1 (the naive ratio 0 / 0 is undefined). According to CoCiP
    conventions, terminus waypoints are "discontinuous" within the flight trajectory, and will not
    contribute to contrail calculations.

    More broadly, any undefined (nan values, or division by 0) segment ratio is set to 1.
    This convention ensures that the contrail calculations are not affected by undefined
    segment-based properties.

    Presently, the output of this function is only used by :func:`plume_temporal_evolution`
    and :func:`new_ice_particle_number` as a scaling term.

    A `seg_ratio` value of 1 is the same as not applying any scaling in these two functions.
    """
    is_defined = (seg_length_t2 > 0.0) & np.isfinite(seg_length_t1)
    default_value = np.ones_like(seg_length_t1)
    return np.divide(seg_length_t1, seg_length_t2, out=default_value, where=is_defined)


def plume_temporal_evolution(
    width_t1: npt.NDArray[np.float_],
    depth_t1: npt.NDArray[np.float_],
    sigma_yz_t1: npt.NDArray[np.float_],
    dsn_dz_t1: npt.NDArray[np.float_],
    diffuse_h_t1: npt.NDArray[np.float_],
    diffuse_v_t1: npt.NDArray[np.float_],
    seg_ratio: npt.NDArray[np.float_] | float,
    dt: npt.NDArray[np.timedelta64] | np.timedelta64,
    max_contrail_depth: float,
) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    """
    Calculate the temporal evolution of the contrail plume parameters.

    Refer to equation (6) of Schumann (2012). See also equations (29), (30), and (31).

    Parameters
    ----------
    width_t1 : npt.NDArray[np.float_]
        contrail width at the start of the time step, [:math:`m`]
    depth_t1 : npt.NDArray[np.float_]
        contrail depth at the start of the time step, [:math:`m`]
    sigma_yz_t1 : npt.NDArray[np.float_]
        sigma_yz governs the contrail plume's temporal evolution at the start of the time step
    dsn_dz_t1 : npt.NDArray[np.float_]
        vertical gradient of the horizontal velocity (wind shear) normal to the contrail axis
        at the start of the time step, [:math:`m s^{-1} / Pa`]::


                           X-----------------------X               X
                                        ^                           |
                                        | (dsn_dz)                  |  <-- (dsn_dz)
                                        |                           |
                                                                    X

    diffuse_h_t1 : npt.NDArray[np.float_]
        horizontal diffusivity at the start of the time step, [:math:`m^{2} s^{-1}`]
    diffuse_v_t1 : npt.NDArray[np.float_]
        vertical diffusivity at the start of the time step, [:math:`m^{2} s^{-1}`]
    seg_ratio : npt.NDArray[np.float_] | float
        Segment length ratio before and after it is advected to the new location.
        See :func:`segment_length_ratio`.
    dt : npt.NDArray[np.timedelta64] | np.timedelta64
        integrate contrails with time steps of dt, [:math:`s`]
    max_contrail_depth: float
        Constrain maximum contrail depth to prevent unrealistic values, [:math:`m`]

    Returns
    -------
    sigma_yy_t2 : npt.NDArray[np.float_]
        The ``yy`` component of convariance matrix, [:math:`m^{2}`]
    sigma_zz_t2 : npt.NDArray[np.float_]
        The ``zz`` component of convariance matrix, [:math:`m^{2}`]
    sigma_yz_t2 : npt.NDArray[np.float_]
        The ``yz`` component of convariance matrix, [:math:`m^{2}`]
    """
    # Convert dt to seconds value and use dtype of other variables
    dtype = np.result_type(width_t1, depth_t1, sigma_yz_t1, dsn_dz_t1, diffuse_h_t1, diffuse_v_t1)
    dt_s = dt / np.timedelta64(1, "s")
    dt_s = dt_s.astype(dtype)

    sigma_yy = 0.125 * (width_t1**2)
    sigma_zz = 0.125 * (depth_t1**2)

    # Convert from max_contrail_depth to an upper bound for diffuse_v_t1
    # All three terms involve the diffuse_v_t1 variable, so we need to
    # calculate the max value for diffuse_v_t1 and apply it to all three terms.
    # If we don't do this, we violate the some mathematical constraints of the
    # covariance matrix (positive definite). In particular, for downstream
    # calculations, we required that
    #   sigma_yy_t2 * sigma_zz_t2 - sigma_yz_t2**2 >= 0
    max_sigma_zz = 0.125 * max_contrail_depth**2
    max_diffuse_v = (max_sigma_zz - sigma_zz) / (2 * dt_s)
    diffuse_v_t1 = np.minimum(diffuse_v_t1, max_diffuse_v)

    # Avoid some redundant calculations
    dsn_dz_t1_2 = dsn_dz_t1**2
    dt_s_2 = dt_s**2
    dt_s_3 = dt_s * dt_s_2

    # Calculate the return arrays
    sigma_yy_t2 = (
        ((2 / 3) * dsn_dz_t1_2 * diffuse_v_t1 * dt_s_3)
        + (dsn_dz_t1_2 * sigma_zz * dt_s_2)
        + (2 * (diffuse_h_t1 + dsn_dz_t1 * sigma_yz_t1) * dt_s)
        + sigma_yy
    ) * (seg_ratio**2)

    sigma_zz_t2 = (2 * diffuse_v_t1 * dt_s) + sigma_zz

    sigma_yz_t2 = (
        (dsn_dz_t1 * diffuse_v_t1 * dt_s_2) + (dsn_dz_t1 * sigma_zz * dt_s) + sigma_yz_t1
    ) * seg_ratio

    return sigma_yy_t2, sigma_zz_t2, sigma_yz_t2


def new_contrail_dimensions(
    sigma_yy_t2: npt.NDArray[np.float_],
    sigma_zz_t2: npt.NDArray[np.float_],
) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    """
    Calculate the new contrail width and depth.

    Parameters
    ----------
    sigma_yy_t2 : npt.NDArray[np.float_]
        element yy, covariance matrix of the Gaussian concentration
        field, Eq. (6) of Schumann (2012)
    sigma_zz_t2 : npt.NDArray[np.float_]
        element zz, covariance matrix of the Gaussian concentration
        field, Eq. (6) of Schumann (2012)

    Returns
    -------
    width_t2 : npt.NDArray[np.float_]
        Contrail width at the end of the time step, [:math:`m`]
    depth_t2 : npt.NDArray[np.float_]
        Contrail depth at the end of the time step, [:math:`m`]
    """
    width_t2 = (8 * sigma_yy_t2) ** 0.5
    depth_t2 = (8 * sigma_zz_t2) ** 0.5
    return width_t2, depth_t2


def new_effective_area_from_sigma(
    sigma_yy: npt.NDArray[np.float_],
    sigma_zz: npt.NDArray[np.float_],
    sigma_yz: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate effective cross-sectional area of contrail plume (``area_eff``) from sigma parameters.

    This method calculates the same output as :func`plume_effective_cross_sectional_area`, but
    calculated with different input parameters.

    Parameters
    ----------
    sigma_yy : npt.NDArray[np.float_]
        element yy, covariance matrix of the Gaussian concentration
        field, Eq. (6) of Schumann (2012)
    sigma_zz : npt.NDArray[np.float_]
        element zz, covariance matrix of the Gaussian concentration
        field, Eq. (6) of Schumann (2012)
    sigma_yz : npt.NDArray[np.float_]
        element yz, covariance matrix of the Gaussian concentration
        field, Eq. (6) of Schumann (2012)

    Returns
    -------
    npt.NDArray[np.float_]
        Effective cross-sectional area of the contrail plume (area_eff)
    """
    det_sigma = sigma_yy * sigma_zz - sigma_yz**2
    return 2 * np.pi * det_sigma**0.5


def new_ice_water_content(
    iwc_t1: npt.NDArray[np.float_],
    q_t1: npt.NDArray[np.float_],
    q_t2: npt.NDArray[np.float_],
    q_sat_t1: npt.NDArray[np.float_],
    q_sat_t2: npt.NDArray[np.float_],
    mass_plume_t1: npt.NDArray[np.float_],
    mass_plume_t2: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Calculate the new contrail ice water content after the time integration step (``iwc_t2``).

    Parameters
    ----------
    iwc_t1 : npt.NDArray[np.float_]
        contrail ice water content, i.e., contrail ice mass per kg of air,
        at the start of the time step, [:math:`kg_{H_{2}O}/kg_{air}`]
    q_t1 : npt.NDArray[np.float_]
        specific humidity for each waypoint at the start of the
        time step, [:math:`kg_{H_{2}O}/kg_{air}`]
    q_t2 : npt.NDArray[np.float_]
        specific humidity for each waypoint at the end of the
        time step, [:math:`kg_{H_{2}O}/kg_{air}`]
    q_sat_t1 : npt.NDArray[np.float_]
        saturation humidity for each waypoint at the start of the
        time step, [:math:`kg_{H_{2}O}/kg_{air}`]
    q_sat_t2 : npt.NDArray[np.float_]
        saturation humidity for each waypoint at the end of the
        time step, [:math:`kg_{H_{2}O}/kg_{air}`]
    mass_plume_t1 : npt.NDArray[np.float_]
        contrail plume mass per unit length at the start of the
        time step, [:math:`kg_{air} m^{-1}`]
    mass_plume_t2 : npt.NDArray[np.float_]
        contrail plume mass per unit length at the end of the
        time step, [:math:`kg_{air} m^{-1}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Contrail ice water content at the end of the time step, [:math:`kg_{ice} kg_{air}^{-1}`]

    Notes
    -----
    (1) The ice water content is fully conservative.
    (2) ``mass_h2o_t2``: the total H2O mass (ice + vapour) per unit of
        contrail plume [Units of kg-H2O/m]
    (3) ``q_sat`` is used to calculate mass_h2o because air inside the
        contrail is assumed to be ice saturated.
    (4) ``(mass_plume_t2 - mass_plume) * q_mean``: contrail absorbs
        (releases) H2O from (to) surrounding air.
    (5) ``iwc_t2 = mass_h2o_t2 / mass_plume_t2 - q_sat_t2``: H2O in the
        gas phase is removed (``- q_sat_t2``).
    """
    q_mean = 0.5 * (q_t1 + q_t2)
    mass_h2o_t1 = mass_plume_t1 * (iwc_t1 + q_sat_t1)
    mass_h2o_t2 = mass_h2o_t1 + (mass_plume_t2 - mass_plume_t1) * q_mean
    iwc_t2 = (mass_h2o_t2 / mass_plume_t2) - q_sat_t2
    return np.maximum(iwc_t2, 0)


def new_ice_particle_number(
    n_ice_per_m_t1: npt.NDArray[np.float_],
    dn_dt_agg: npt.NDArray[np.float_],
    dn_dt_turb: npt.NDArray[np.float_],
    seg_ratio: npt.NDArray[np.float_] | float,
    dt: npt.NDArray[np.timedelta64] | np.timedelta64,
) -> npt.NDArray[np.float_]:
    """Calculate the number of ice particles per distance at the end of the time step.

    Parameters
    ----------
    n_ice_per_m_t1 : npt.NDArray[np.float_]
        number of contrail ice particles per distance at the start of
        the time step, [:math:`m^{-1}`]
    dn_dt_agg : npt.NDArray[np.float_]
        rate of ice particle losses due to sedimentation-induced aggregation, [:math:`# s^{-1}`]
    dn_dt_turb : npt.NDArray[np.float_]
        rate of contrail ice particle losses due to plume-internal turbulence, [:math:`# s^{-1}`]
    seg_ratio : npt.NDArray[np.float_] | float
        Segment length ratio before and after it is advected to the new location.
    dt : npt.NDArray[np.timedelta64] | np.timedelta64
        integrate contrails with time steps of dt, [:math:`s`]

    Returns
    -------
    npt.NDArray[np.float_]
        number of ice particles per distance at the end of the time step, [:math:`m^{-1}`]
    """
    # Convert dt to seconds value and use dtype of other variables
    dtype = np.result_type(n_ice_per_m_t1, dn_dt_agg, dn_dt_turb, seg_ratio)
    dt_s = dt / np.timedelta64(1, "s")
    dt_s = dt_s.astype(dtype)

    n_ice_per_m_t1 = np.maximum(n_ice_per_m_t1, 0)

    exp_term = np.where(dn_dt_turb * dt_s < 80, np.exp(-dn_dt_turb * dt_s), 0)

    numerator = dn_dt_turb * n_ice_per_m_t1 * exp_term
    denominator = dn_dt_turb + (dn_dt_agg * n_ice_per_m_t1 * (1 - exp_term))
    n_ice_per_m_t2 = (numerator / denominator) * seg_ratio

    small_loss = (dn_dt_turb * dt_s) < 1e-5  # For small ice particle losses
    denom = 1 + (dn_dt_agg * dt_s * n_ice_per_m_t1)
    n_ice_per_m_t2[small_loss] = n_ice_per_m_t1[small_loss] / denom[small_loss]
    return np.maximum(n_ice_per_m_t2, 1)


########
# Energy Forcing
########
# TODO: This should be moved closer to the radiative forcing calculations


def energy_forcing(
    rf_net_t1: npt.NDArray[np.float_],
    rf_net_t2: npt.NDArray[np.float_],
    width_t1: npt.NDArray[np.float_],
    width_t2: npt.NDArray[np.float_],
    seg_length_t2: npt.NDArray[np.float_] | float,
    dt: npt.NDArray[np.timedelta64] | np.timedelta64,
) -> npt.NDArray[np.float_]:
    """Calculate the contrail energy forcing over time step.

    The contrail energy forcing is calculated as the local contrail net
    radiative forcing (RF', change in energy flux per contrail area) multiplied
    by its width and integrated over its length and lifetime.

    Parameters
    ----------
    rf_net_t1 : npt.NDArray[np.float_]
        local contrail net radiative forcing at the start of the time step, [:math:`W m^{-2}`]
    rf_net_t2 : npt.NDArray[np.float_]
        local contrail net radiative forcing at the end of the time step, [:math:`W m^{-2}`]
    width_t1 : npt.NDArray[np.float_]
        contrail width at the start of the time step, [:math:`m`]
    width_t2 : npt.NDArray[np.float_]
        contrail width at the end of the time step, [:math:`m`]
    seg_length_t2 : npt.NDArray[np.float_] | float
        Segment length of contrail waypoint at the end of the time step, [:math:`m`]
    dt : npt.NDArray[np.timedelta64] | np.timedelta64
        integrate contrails with time steps of dt, [:math:`s`]

    Returns
    -------
    npt.NDArray[np.float_]
        Contrail energy forcing over time step dt, [:math:`J`].
    """
    rad_flux_per_m = mean_radiative_flux_per_m(rf_net_t1, rf_net_t2, width_t1, width_t2)
    energy_flux_per_m = mean_energy_flux_per_m(rad_flux_per_m, dt)
    return energy_flux_per_m * seg_length_t2


def mean_radiative_flux_per_m(
    rf_net_t1: npt.NDArray[np.float_],
    rf_net_t2: npt.NDArray[np.float_],
    width_t1: npt.NDArray[np.float_],
    width_t2: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """Calculate the mean radiative flux per length of contrail between two time steps.

    Parameters
    ----------
    rf_net_t1 : npt.NDArray[np.float_]
        local contrail net radiative forcing at the start of the time step, [:math:`W m^{-2}`]
    rf_net_t2 : npt.NDArray[np.float_]
        local contrail net radiative forcing at the end of the time step, [:math:`W m^{-2}`]
    width_t1 : npt.NDArray[np.float_]
        contrail width at the start of the time step, [:math:`m`]
    width_t2 : npt.NDArray[np.float_]
        contrail width at the end of the time step, [:math:`m`]

    Returns
    -------
    npt.NDArray[np.float_]
        Mean radiative flux between time steps, [:math:`W m^{-1}`]
    """
    rad_flux_per_m_t1 = width_t1 * rf_net_t1
    rad_flux_per_m_t2 = width_t2 * rf_net_t2
    return (rad_flux_per_m_t1 + rad_flux_per_m_t2) * 0.5


def mean_energy_flux_per_m(
    rad_flux_per_m: npt.NDArray[np.float_], dt: npt.NDArray[np.timedelta64] | np.timedelta64
) -> npt.NDArray[np.float_]:
    """Calculate the mean energy flux per length of contrail on segment following waypoint.

    Parameters
    ----------
    rad_flux_per_m : npt.NDArray[np.float_]
        Mean radiative flux between time steps for waypoint, [:math:`W m^{-1}`].
        See :func:`mean_radiative_flux_per_m`.
    dt : npt.NDArray[np.timedelta64]
        timedelta of integration timestep for each waypoint.

    Returns
    -------
    npt.NDArray[np.float_]
        Mean energy flux per length of contrail after waypoint, [:math:`J m^{-1}`]

    Notes
    -----
    Implementation differs from original fortran in two ways:

    - Discontinuity is no longer set to 0 (this occurs directly in model :class:`Cocip`)
    - Instead of taking an average of the previous and following segments,
      energy flux is only calculated for the following segment.

    See Also
    --------
    :func:`mean_radiative_flux_per_m`
    """
    dt_s = dt / np.timedelta64(1, "s")
    dt_s = dt_s.astype(rad_flux_per_m.dtype)
    return rad_flux_per_m * dt_s
